function [ output_args ] =com(config_file, num_fext, num_next, varargin)
% For use in the development of study group, task force, and ad hoc documents
% This is not a normative or an official IEEE document
% This code is expected to change as the document evolves
% $Id: com_ieee8023_93a.m.rca 1.17 Thu Jun 20 10:19:39 2013 aran Experimental aran $
% See bottom of this file for detailed change log.

% structures:
% chdata(i)
%           i= 1 --> THRU index
%           i= 2, num_fext+1 --> FEXT channel index
%           i= num_fext+2, num_next+num_fext+1
%       base: name of THRU file
%          A: amplitude
%       type: 'THRU', 'NEXT', or 'FEXT'
%        ftr: Rise time frequency
%      fmaxi: max number of frequency points
%      faxis: frequency array
%      sdd21: (Htot) rewritten as product of ,vtf based on pkg RL ,tx filter, Rx filter
%      sdd22: differential RL
%      sdd11: differential RL
%     sdd21p: vtf based on pkg RL , this an interim parameter and set to sdd21
%     sdd21f: raw differential IL not filtered use for FD plots


%% SECTION TITLE: file_setup
% DESCRIPTIVE TEXT acquire parsing command string and set up OP control structure. Then read in files
close all force
display('COM example relecting IEEE 802.3 Annex 93A, Draft 2.1($Revision: 1.17 $)')
display(sprintf('\tThis code is expected to change as the document evolves. \n\tThis is not a normative or an official IEEE document.'))
% need to see what happens for version 8
OP.VEROK=verLessThan('matlab', '7.4.1');
if OP.VEROK; display('Matlab version 7.4 or higher required'); return; end
% get the first 3 arguments and allow for interactive input.
if ~exist('config_file','var')
    config_file=input('Enter config XLS file or return will just pop a window to ask for the XLS file]:  ','s') ;
end
param.exist=1; % Just defines the base structure param
% get config file, test if passed, if not pop window to ask
if strfind(upper(config_file),'.XLS') ~= 0
    file = config_file;
    [param OP] = read_ParamConfigFile(file, param, OP);
else % ask for config file
    [file, UNUSED_OUTPUT] = uigetfile({'*.xls'},'INPUT CONFIG FILE .xls');
    [param OP] = read_ParamConfigFile(file, param, OP);
end

if ~exist('num_fext','var')
    if OP.RX_CALIBRATION
        num_fext=1;
        display('First prompt is for the measured test thru channel and following prompt is for Rx noise path channel')
    else
        num_fext=input('How many FEXT channels are to be entered? [return is means no FEXT] ');
    end
    if isempty(num_fext)==1, num_fext=0; end
end
if ~exist('num_next','var')
    if OP.RX_CALIBRATION
        num_next=0;
    else
        num_next=input('How many NEXT channels are to be entered? [return is means no NEXT] ');
    end
    if isempty(num_next)==1, num_next=0; end
end
xtk=num_fext+num_next; % total number of crosstalk aggressors
param.num_next=num_next;
param.num_fext=num_fext;
param.K=num_fext+num_next+1;
% checking for data when running for rx compliance BBN calibration
if OP.RX_CALIBRATION == 1
    if num_fext ~=1
        h = msgbox(['One and only noise path channel is required ']); set(h,'Color',[1 .85 0]);
        movegui(h,'northwest')
        if OP.DEBUG ~= 1
            return
        end
    end
    h = msgbox(['Please make sure the measured "sigma_rj" and "Add" fields in the config xls file have been modified from the Tx measurement. ']); set(h,'Color',[0 1 1]);
    movegui(h,'southeast')
end
% create result directory if needed
if ~exist(OP.RESULT_DIR,'dir'); mkdir(OP.RESULT_DIR); end
OP.BMXLMT=1; % set to 1 to limit dfe tap 1 to 1
% allow finite impulse response input rather that s-parameters with
% OP.EXTERNAL = true. However the use_external_IR function is not provided
OP.EXTERNAL = false;
if size(varargin,2) ~= 0 % process case where file names are passed in function call
    if strfind(upper(char(varargin(1))),'EXTERNAL_IR') ~= 0
        OP.EXTERNAL = true;
        OP.GET_FD = 0;
        ir1a= varargin(2);
        ex_var = varargin(3);
        [chdata OP param ]  = use_external_IR(param, OP ,num_fext,num_next,0,ir1a,ex_var);
        kxi=num_fext+num_next+1;
    else
        OP.EXTERNAL = false;
        if xtk +1 ~= size(varargin,2)% check that number of varargin arguments passed becomes is number of crosstalk files+1 (thru)
            display('total files must be next + fext + a thru')
            return
        end
    end
end
if OP.EXTERNAL == false
    % filename parsing and acquisition
    %------------------------------------------------------------------
    %----------put files names into chdata structure ---------
    % The thru file has the index of 1
    % crosstalk file are indexed from 2
    % nxi is incremented each time a file is read in  so that nxi will end
    filepath=[]; % path name for file
    nxi=0; % file index
    % get the THRU file
    if size(varargin,2) ~= 0
        varargin(1)=strrep(varargin(1),'\', filesep); % OS file convention conversion
        [filepath, basename, fileext]=fileparts(varargin{1});
        
    else
        if OP.RX_CALIBRATION == 1
            h = msgbox(['enter test channel s-parameter file']); set(h,'Color',[1 .85 0]);
            movegui(h,'northeast')
        end
        dir=fullfile(filepath, '*.s4p');
        [basename,filepath]=uigetfile(dir,'input thru channel response .s4p ');
        if filepath == 0
            error('No Thru file')
        end
        [UNUSED_OUTPUT, basename, fileext]=fileparts(basename);
    end
    nxi=nxi+1;
    filename{nxi} = fullfile(filepath,  [basename fileext]);
    ext{nxi} = fileext;
    [pth, dirname]=fileparts(filepath);   
%    chdata(nxi).base=[pth(max(strfind(pth,filesep))+1:end) '--' basename]; % add 1 directory back to basename
    chdata(nxi).base=[dirname '--' basename]; % add 1 directory back to basename
    chdata(nxi).A=param.a_thru; % pam encoding amplitude reduction is do in reporting but is comprehending in crosstalk PDF
    chdata(nxi).type='THRU'; % thru type
    chdata(nxi).ftr=param.fb*param.f_v;
    param.base=chdata(nxi).base; %for print out function that don't pass chdata
    
    % now get FEXT file names into chdata structure
    kxi=nxi;
    for nxi=kxi+1:num_fext+kxi
        lastfilepath=filepath;
        if size(varargin,2) ~= 0
            [filepath, basename, fileext]=fileparts(varargin{nxi});
        else
            if OP.RX_CALIBRATION == 1
                h = msgbox(['enter noise channel s-parameter file']); set(h,'Color',[1 .85 0]);
                movegui(h,'northeast')
            end
            dir=fullfile(filepath, '*.s4p');
            if OP.RX_CALIBRATION == 1
                [basename,filepath]=uigetfile(dir,['input noise channel response .s4p']);
            else
                [basename,filepath]=uigetfile(dir,['input fext channel response .s4p #', num2str(nxi-kxi)]);
            end
            if filepath==0
                error('Not enough NEXT files')
            end
            [UNUSED_OUTPUT, basename, fileext]=fileparts(basename);
        end
        if isempty( filepath), filepath=lastfilepath; end
        filename{nxi} = fullfile(filepath,  [basename fileext]);
        ext{nxi} = fileext;
        [UNUSED_OUTPUT, dirname]=fileparts(filepath);
        chdata(nxi).base=[dirname '--' basename];
        chdata(nxi).A=param.a_fext;
        chdata(nxi).ftr=param.fb*param.f_f;
        chdata(nxi).type='FEXT'; % fext type
    end
    % now get NEXT file names into chdata structure
    kxi=num_fext+kxi;
    for nxi=kxi+1:num_next+kxi
        lastfilepath=filepath;
        if size(varargin,2) ~= 0
            [filepath, basename, fileext]=fileparts(varargin{nxi});
        else
            dir=fullfile(filepath, '*.s4p');
            [basename,filepath]=uigetfile(dir,['input next channel response .s4p ', num2str(nxi-kxi)]);
            if filepath==0
                error('Not enough NEXT files')
            end
            [UNUSED_OUTPUT, basename, fileext]=fileparts(basename);
        end
        if isempty( filepath), filepath=lastfilepath; end
        filename{nxi} = fullfile(filepath,  [basename fileext]);
        ext{nxi} = fileext;
        [UNUSED_OUTPUT, dirname]=fileparts(filepath);
        chdata(nxi).base=[dirname '--' basename];
        chdata(nxi).A=param.a_next;
        chdata(nxi).ftr=param.fb*param.f_n;
        chdata(nxi).type='NEXT'; % next type
    end
    
    %% SECTION TITLE: extract s-parameter and convert to differential mode
    % DESCRIPTIVE TEXT extract s-parameter data from files and apply tx and rx filters as well as package filters
    kxi=num_next+kxi;
    if ~OP.DISPLAY_WINDOW, fprintf('reading file '); end
    for i=1:kxi
        if OP.DISPLAY_WINDOW; hwaitbar=waitbar(0);end
        progress = (i-1)/kxi;
        if OP.DISPLAY_WINDOW
            waitbar(progress, hwaitbar, 'reading files'); drawnow;
        else
            fprintf(['%i '],i);
        end
        
        switch ext{i}
            case '.s4p'
                if ~isfield(param,'snpPortsOrder') || isempty(param.snpPortsOrder)
                    param.snpPortsOrder = [1 3 2 4]; % default order normally used.
                elseif length(param.snpPortsOrder) ~= 4
                    error( 'warning:sNpFilePortMismatch', ...
                        '\n\t The number of ports defined (%G) does not match the sNp file type (%s)', ...
                        length(param.snpPortsOrder), ...
                        ext{i,:} ...
                        );
                end
                % read function returns differential mode parameters
                [Sch SDDch] = read_p4_s4params(filename{i},  0, 0, param.snpPortsOrder, OP);
                chdata(i).fmaxi = length(Sch.freq);
                
                %%%%Warnings for maximum frequency provided and frequency step%%%%%%%%%%%%%%%%%%%%%%%%%%
                if Sch.freq(chdata(i).fmaxi) < param.fb
                    warning('The maximum frequency provided, %g, is less than the signaling rate: %g', ...
                        Sch.freq(end), param.fb);
                end
                freqstep=diff(Sch.freq);
                if std(freqstep)~=0
                    warning('non-equally-spaced frequency: min=%g, max=%g', ...
                        min(freqstep), max(freqstep));
                end
                if max(freqstep) > 0.01e9
                    warning('Frequency step, %g, is larger than the recommended 0.01 GHz', max(freqstep));
                end
                %%%%Warnings for maximum frequency provided and frequency step%%%%%%%%%%%%%%%%%%%%%%%%%%
                
                chdata(i).faxis = Sch.freq;
                chdata(i).sdd12 = transpose(SDDch(1:chdata(i).fmaxi,1,2));
                chdata(i).sdd21 = transpose(SDDch(1:chdata(i).fmaxi,2,1));
                chdata(i).sdd22 = transpose(SDDch(1:chdata(i).fmaxi,2,2));
                chdata(i).sdd11 = transpose(SDDch(1:chdata(i).fmaxi,1,1));
                
            case '.s12p' % not fully implemented, this is for a 3 pair model only
                if ~isfield(param,'snpPortsOrder') || isempty(param.snpPortsOrder)
                    param.snpPortsOrder = [1 2 3 4 6 7 8 9 10 11 12]; % default order normally used.
                elseif length(param.snpPortsOrder) ~= 12
                    error( 'warning:sNpFilePortMismatch', ...
                        '\n\t The number of ports defined (%G) does not match the sNp file type (%s)', ...
                        length(param.snpPortsOrder), ...
                        ext{i,:} ...
                        );
                end
                Sch = readdataSnPx(filename(i,:), 12);
                chdata(i).fmaxi = length(Sch.freq);
                chdata(i).faxis = Sch.freq;
                if isequal([1 3 5 7 9 11 2 4 6 8 10 12], param.snpPortsOrder)
                    chdata(i).sdd21 = reduce((Sch.cs(6,5,1:chdata(i).fmaxi)-Sch.cs(8,5,1:chdata(i).fmaxi)-Sch.cs(6,7,1:chdata(i).fmaxi)+Sch.cs(8,7,1:chdata(i).fmaxi))./2); %dd21=s21-s41-s23+s43
                    chdata(i).sdd11 = reduce((Sch.cs(5,5,1:chdata(i).fmaxi)-Sch.cs(7,5,1:chdata(i).fmaxi)-Sch.cs(5,7,1:chdata(i).fmaxi)+Sch.cs(7,7,1:chdata(i).fmaxi))./2); %dd11=s11-s31-s13+s33
                    chdata(i).sdd22 = reduce((Sch.cs(6,6,1:chdata(i).fmaxi)-Sch.cs(6,8,1:chdata(i).fmaxi)-Sch.cs(8,6,1:chdata(i).fmaxi)+Sch.cs(8,8,1:chdata(i).fmaxi))./2); %dd22=s22-s41-s14+s44
                elseif isequal([1 2 3 4 5 6 7 8 9 10 11 12], param.snpPortsOrder)
                    chdata(i).sdd21 = reduce((Sch.cs(9,3,1:chdata(i).fmaxi)-Sch.cs(10,3,1:chdata(i).fmaxi)-Sch.cs(9,4,1:chdata(i).fmaxi)+Sch.cs(10,4,1:chdata(i).fmaxi))./2);  %dd21=s21-s41-s23+s43
                    chdata(i).sdd11 = reduce((Sch.cs(3,3,1:chdata(i).fmaxi)-Sch.cs(4,9,1:chdata(i).fmaxi)-Sch.cs(3,4,1:chdata(i).fmaxi)+Sch.cs(4,4,1:chdata(i).fmaxi) )/2);    %dd11=s11-s31-s13+s33
                    chdata(i).sdd22 = reduce((Sch.cs(9,9,1:chdata(i).fmaxi)-Sch.cs(10,3,1:chdata(i).fmaxi)-Sch.cs(9,10,1:chdata(i).fmaxi)+Sch.cs(10,10,1:chdata(i).fmaxi) )/2);%dd22=s22-s41-s14+s44
                else
                    display('INVALID ENTRY FOR TX PORTS.  Use [1 2 3 4 5 6] or [1 3 5 7 9 11]')
                    return
                end
            otherwise
        end
        %
        % differential response is sdd21 <------------------------------
        % differential return losses are sdd11 and sdd22
        if isempty(strfind(param.type,'CLAUSE 92')) == 0
            % add boards to sdd
            [chdata(i).sdd11, chdata(i).sdd12,chdata(i).sdd21, chdata(i).sdd22] =add_brd(chdata, i, param);
            end
        % the new sdd
        % changes to s21_pkg for calibration is to not include Tx package in the sp21p
        chdata(i).sdd21p= s21_pkg(chdata(i).sdd21, chdata(i).sdd11, chdata(i).sdd22, chdata(i).faxis, param, OP, chdata(i).type, i);
        chdata(i).sdd21f=chdata(i).sdd21; % used for FD analysis i.e. not filtered
        if  OP.INC_PACKAGE ~= 0,
            chdata(i).sdd21=chdata(i).sdd21p;
        end
        % add in the Tx and Rx filter option add and addition Rx
        % package loss
        f_r=param.f_r*param.fb;
        param.H_r = 1./polyval([1 2.613126 3.414214 2.613126 1], 1i*chdata(i).faxis./f_r);
        if OP.INCLUDE_FILTER ~=0
            chdata(i).sdd21=chdata(i).sdd21.*param.H_r;
        end
    end
    if ~OP.DISPLAY_WINDOW, fprintf('\n'); end
end
%% FD processing s-parameter
% at this point sdd21 responses and faxis (frequency) array are defined
ICN=0;
for i=1:kxi
    if OP.GET_FD~=0
        if i == 2, PSXT(1:length(chdata(i).sdd21f))=0;end
        a=find(chdata(i).faxis(:)>param.f2);
        if isempty(a)
            param.f2=chdata(i).faxis(length(chdata(i).faxis));
            index_f2=length(chdata(i).faxis);
        else
            index_f2=a(1);
        end
        % R is the frequency dependent parameter for the sinc function use in the
        % PWF for ICN
        temp_angle=(param.samples_per_ui*param.sample_dt)*pi.*chdata(i).faxis;
        if(chdata(i).faxis(1)==0)
            temp_angle(1)=1e-20;% we don't want to divide by zero
        end
        SINC = sin(temp_angle)./temp_angle;
        PWF_data=SINC.^2;
        PWF_trf=(1+(chdata(i).faxis/chdata(i).ftr).^4).^-1;
        %// bw1=2.613126; bw2=3.4142136; bw3=2.613126;
        fr=param.f_r*param.fb;
        PWF_rx=(1+(chdata(i).faxis/fr).^8).^-1;
        %// PWF_rx=1./(1+(chdata(i).faxis./(param.fb*.75)).^8); % receiver filter
        PWF_highpass=1;
        PWF=PWF_data.*PWF_trf.*PWF_rx.*PWF_highpass; % power weight function
        % freq delta for integration
        chdata(i).delta_f=chdata(i).faxis(11)-chdata(i).faxis(10);
        % from ba spec, this is basically ICN
        chdata(i).sigma_FD=sqrt(2*chdata(i).delta_f/param.fb*sum( PWF(1:index_f2).*abs(chdata(i).sdd21f(1:index_f2)).^2))/2;
        if i==1
            [chdata(i).iln chdata(i).fit] = get_ILN(chdata(i).sdd21f(1:index_f2), chdata(i).faxis(1:index_f2));
            nq_indx=find(chdata(i).faxis >= 1/param.ui/2,1, 'first');
            chdata(i).fit_ILatNq=-20*log10(abs(chdata(i).fit(nq_indx)));
            chdata(i).ILatNq=-20*log10(abs(chdata(i).sdd21f(nq_indx)));
            fit_loss=chdata(i).fit_ILatNq;
            Nq_loss=chdata(i).ILatNq;
            ILD=20*log10(abs(chdata(i).fit))-20*log10(abs(chdata(i).sdd21f(1:index_f2)));
            ILD_RMS=sqrt(2*chdata(i).delta_f/param.fb*sum( PWF(1:index_f2).*ILD(1:index_f2).^2));
            if OP.DEBUG
                display(['Insertion Loss at Nyquist = ', num2str(chdata(i).ILatNq)])
            end
            if OP.DISPLAY_WINDOW
                figure(300)
                subplot(3,1,1)
                title('Losses')
                plot(chdata(i).faxis,20*log10(abs(chdata(i).sdd21f)),'Disp','IL')
                hold on
                plot(chdata(i).faxis(1:index_f2),20*log10(abs(chdata(i).fit(1:index_f2))),'g','Disp','IL fit')
                plot(chdata(i).faxis,20*log10(abs(chdata(i).sdd11)),'c','Disp','RL11')
                plot(chdata(i).faxis,20*log10(abs(chdata(i).sdd22)),'- c','Disp','RL22')
                hold off
                subplot(3,1,3)
                plot(chdata(i).faxis(1:index_f2),ILD,'Disp','ILD')
            end
        else
            PSXT=sqrt((abs(chdata(i).sdd21f)).^2+PSXT.^2); % power sum xtk
            ICN=sqrt(2*chdata(i).delta_f/param.fb*sum( PWF(1:index_f2).*abs(PSXT(1:index_f2)).^2))/2;
        end
    end
end
%%
% at this point sdd21 responses and faxis (frequency) array are defined
% ----------
% first find equalization settings
%            When a thru channel is give and it should be unless we are in a
%            diagnostic mode
%
%
%ICN=0;
if OP.RX_CALIBRATION ~= 0, kxi=1;xtk=0 ;  end
for i=1:kxi
    if OP.GET_FD~=0
        if i == 2, PSXT(1:length(chdata(i).sdd21f))=0;end
        a=find(chdata(i).faxis(:)>param.f2);
        if isempty(a)
            param.f2=chdata(i).faxis(length(chdata(i).faxis));
            index_f2=length(chdata(i).faxis);
        else
            index_f2=a(1);
        end
        % R is the frequency dependent parameter for the sinc function use in the
        % PWF for ICN
        temp_angle=(param.samples_per_ui*param.sample_dt)*pi.*chdata(i).faxis;
        if(chdata(i).faxis(1)==0)
            temp_angle(1)=1e-20;% we don't want to divide by zero
        end
        SINC = sin(temp_angle)./temp_angle;
        PWF_data=SINC.^2;
        PWF_trf=(1+(chdata(i).faxis/chdata(i).ftr).^4).^-1;
        %// bw1=2.613126; bw2=3.4142136; bw3=2.613126;
        fr=.75*param.fb;
        PWF_rx=(1+(chdata(i).faxis/fr).^8).^-1;
        %PWF_rx=1./(1+(chdata(i).faxis./(param.fb*.75)).^8); % receiver filter
        PWF_highpass=1;
        PWF=PWF_data.*PWF_trf.*PWF_rx.*PWF_highpass; % power weight function
        % freq delta for integration
        chdata(i).delta_f=chdata(i).faxis(11)-chdata(i).faxis(10);
        % from ba spec, this is basically ICN
        chdata(i).sigma_FD=sqrt(2*chdata(i).delta_f/param.fb*sum( PWF(1:index_f2).*abs(chdata(i).sdd21f(1:index_f2)).^2))/2;
        if i==1
            [chdata(i).iln chdata(i).fit] = get_ILN(chdata(i).sdd21f(1:index_f2), chdata(i).faxis(1:index_f2));
            nq_indx=find(chdata(i).faxis >= 1/param.ui/2,1, 'first');
            chdata(i).fit_ILatNq=-20*log10(abs(chdata(i).fit(nq_indx)));
            chdata(i).ILatNq=-20*log10(abs(chdata(i).sdd21f(nq_indx)));
            fit_loss=chdata(i).fit_ILatNq;
            Nq_loss=chdata(i).ILatNq;
            ILD=20*log10(abs(chdata(i).fit))-20*log10(abs(chdata(i).sdd21f(1:index_f2)));
            ILD_RMS=sqrt(2*chdata(i).delta_f/param.fb*sum( PWF(1:index_f2).*ILD(1:index_f2).^2));
            if OP.DEBUG
                display(['Insertion Loss at Nyquist = ', num2str(chdata(i).ILatNq)])
            end
            if OP.DISPLAY_WINDOW
                figure(300)
                subplot(3,1,1)
                title('Losses')
                plot(chdata(i).faxis,20*log10(abs(chdata(i).sdd21f)),'Disp',['IL '])
                hold on
                plot(chdata(i).faxis(1:index_f2),20*log10(abs(chdata(i).fit)),'g','Disp','IL fit')
                plot(chdata(i).faxis,20*log10(abs(chdata(i).sdd11)),'c','Disp','RL11')
                plot(chdata(i).faxis,20*log10(abs(chdata(i).sdd22)),'- c','Disp','RL22')
                hold off
                subplot(3,1,3)
                plot(chdata(i).faxis(1:index_f2),ILD,'Disp','ILD')
            end
        else
            PSXT=sqrt((abs(chdata(i).sdd21f)).^2+PSXT.^2); % power sum xtk
            ICN=sqrt(2*chdata(i).delta_f/param.fb*sum( PWF(1:index_f2).*abs(PSXT(1:index_f2)).^2))/2;
        end
    end
end
if OP.DISPLAY_WINDOW && OP.GET_FD ~=0
    figure(300)
    if kxi > 1
        scale=param.a_fext/param.a_thru;% added for crosstalk scaling
        subplot(3,1,1)
        hold on
        plot(chdata(i).faxis,20*log10(abs(PSXT*scale)),'r','Disp',['PSXTK'])
        icrxi=find(chdata(i).faxis >=param.fb/2,1,'first');
        subplot(3,1,2)
        grid on
        ILtemp=20*log10(abs(chdata(1).sdd21f));
        IL4ICR=interp1(chdata(1).faxis,ILtemp,chdata(i).faxis);
        scale=param.a_fext/param.a_thru;% added for crosstalk scaling
        ICR=-20*log10(abs(PSXT*scale))+IL4ICR;
        semilogx(chdata(i).faxis,ICR,'Disp',[param.base ' ICR'])
        hold on
        stem(chdata(i).faxis(icrxi),ICR(icrxi),'g')
        hold off
    end
    h1=figure(300);
    set(h1,'Position', [20 20 800 800] );
    subplot(3,1,1)
    title([param.base ' Losses'])
    ylabel('dB')
    xlabel('Hz')
    grid on
    subplot(3,1,2)
    ylim([0 80])
    ylabel('dB')
    xlabel('Hz')
    title([param.base ' ICR'])
    grid on
    subplot(3,1,3)
    ylim([-8 8])
    ylabel('dB')
    xlabel('Hz')
    title([param.base ' ILD'])
    grid on
    hold off
end
COM=1000; % reset COM prior to calibration
param.WGN=0;
DO_ONCE=true;
%% at this point only the impulse responses are needed. However vestiges FD may be intermingled
while (OP.RX_CALIBRATION==1 || DO_ONCE==true) && COM > param.cc1
    DO_ONCE=false;
    % get impulse responses which in interim step between equation for X(f) and
    % H^(k)(t) without TX FFE or CTLE. These will we added later.
    for i=1:kxi
        if i==1
            if OP.EXTERNAL == false
                [chdata(i).ir chdata(i).t] = s21_to_impulse_DC(chdata(i).sdd21 ,chdata(i).faxis,param.sample_dt) ;
                %//                RAWSBR=conv(ones(param.samples_per_ui,1),chdata(i).ir); % reference to 1V
                %//                RAWSBR=RAWSBR(1:end-param.samples_per_ui+1);
                chdata(i).ir=chdata(i).ir*chdata(i).A; % adjust IRx for amplitude
            end
        else
            if OP.EXTERNAL == false
                [chdata(i).ir chdata(i).t] = s21_to_impulse_DC(chdata(i).sdd21 ,chdata(i).faxis,param.sample_dt) ;
                chdata(i).ir=chdata(i).ir*chdata(i).A; % adjust IRx for amplitude
            end
        end
        %------------------------------------------------------------
        % next find Pulse response (SBR) for each channel h^(k)(t)
        if ~OP.DISPLAY_WINDOW && i==1, fprintf('processing COM PDF '); end
        chdata(i).result=get_SBR(chdata(i).ir, param.samples_per_ui);
        if OP.SAVE_RESP ~= 0
            %//            t=chdata(i).t;
            %//            SBR=chdata(i).result.SBR;
            %//            ir=chdata(i).IRx;
            %//            if OP.GET_FD == 1
            %//                IL_fil=chdata(i).sdd21;
            %//                IL=chdata(i).sdd21f;
            %//                f=chdata(i).faxis;
        end
        % plot results
        if OP.DEBUG
            figure(i+150); subplot(2,1,1); set(gcf, 'Name', chdata(i).base);
            plot(chdata(i).t, chdata(i).result.SBR,'Disp',[chdata(i).base ' - PR'])
            title(strrep(chdata(i).base,'_',' '))
            ylabel('Volts')
            xlabel('seconds')
            
            if OP.GET_FD~=0, display([chdata(i).base ' FD rms (mv) =' num2str(1000*chdata(i).sigma_FD) ]); end
            if (param.ndfe ~=0 ), display([ ' Pulse Height = ' num2str(max(chdata(i).result.SBR)) ]); end
        end
    end
end

% Determine equalization settings
[result, UNUSED_OUTPUT, param, chdata]= optimize_fom(OP,param, chdata);

chdata(1).IRx=result.IR; % IRx has the TX FFE and CTLE applied
max_signal=result.A_s; % this is the "s" in SNR (PAM4 gain in handled in last sections
% the pdf for PAM4 uses the full swing SBR but assigns voltage for the PDF accordingly
for i=1:kxi
    % get quick PDF  results
    if isequal(chdata(i).type, 'THRU')
        excluded_UI=param.ndfe;
        if OP.GET_FD ~=0, FD_rms_sci=chdata(i).sigma_FD; end % decide later whether to use FD or TD rms. pdfr has an rms field.
    else
        excluded_UI=0;
    end
    if OP.INCLUDE_CTLE==1; chdata(i).ir  = TD_CTLE(chdata(i).ir,param.fb,result.eq.ctle.maxkacdc,param.samples_per_ui);end
    % need to add in the tx ff3 filter here if FEXT
    if isequal(chdata(i).type, 'FEXT') || isequal(chdata(i).type, 'THRU')
        upsampled_txffe = zeros(1, param.samples_per_ui*2+1); % start with zeros everywhere
        upsampled_txffe(1+(0:2)*param.samples_per_ui) = result.eq.txle; % plant the coefficients in the desired locations
        chdata(i).ir = filter(upsampled_txffe, 1, chdata(i).ir);
    end
    chdata(i).result=get_SBR(chdata(i).ir, param.samples_per_ui);
    % the max signal here comprehends drive strength but no coding
    [pdf] = get_pdf(max_signal, excluded_UI, i, chdata, param, OP) ;
    chdata(i).pdfr=pdf;
    % reporting
    a=find(cumsum(chdata(i).pdfr.y) >1e-12,1,'first');
    chdata(i).maxquickpdf=(chdata(i).pdfr.y(a));
    
end
if ~OP.DISPLAY_WINDOW, fprintf('\n'); end

if OP.RX_CALIBRATION
%//    [ sigma_noise sigma_N ] = get_sigma_noise( result.eq.ctle.maxkacdc, param, chdata );
    [ UNUSED_OUTPUT, sigma_N ] = get_sigma_noise( result.eq.ctle.maxkacdc, param, chdata );
else
    %//         param.sigma_r_mod=param.Na_rms;
    %//         H_ctle=(4i*chdata(1).faxis/param.fb + result.eq.ctle.maxkacdc)./((1+1i*chdata(1).faxis/param.fb).*(1+4i*chdata(1).faxis/param.fb));
    %//         param.sigma_r_mod = sqrt(param.eta_0*sum(abs(param.H_r(2:end) .* H_ctle(2:end)).^2 .* diff(chdata(1).faxis)/1e9));
    sigma_N = result.sigma_N;
end

%% Equation 93A-38 %%
sigma_G = norm([param.sigma_RJ*param.sigma_X*norm(result.h_J), sigma_N]);

%% Equation 93A-39 %%
%sig_depend_noise = normal_dist(sigma_G,7.5,joint_pdf.BinSize);
gaussian_noise_pdf = normal_dist(sigma_G,7.5,param.BinSize);

%% p_DD is computed using the procedure defined in 93A.1.7.1 with h(n)=A_DD*h_J(n)
p_DD = get_pdf_from_sampled_signal(param.A_DD*result.h_J, param.levels, param.BinSize);

%% Equation 93A-40
noise_pdf=conv_fct(gaussian_noise_pdf, p_DD);

%% Implementation of 93A.1.7.3 combination procedure
%%  (effectively Equation 93A-41) %%

% Self-Channel Interference is thru residual result
sci_pdf = chdata(1).pdfr;
sci_mxi=find(cumsum(sci_pdf.y)>=param.specBER, 1, 'first');
thru_max_peak_interference_at_BER=abs(sci_pdf.x(sci_mxi));
sci_msi=find(cumsum(sci_pdf.y)>=1e-12, 1, 'first');
sci_sigma=abs(sci_pdf.x(sci_msi)/(erfcinv(2*1e-12)*sqrt(2)));

% Co-Channel Interference PDFs (for information only):
% initialize to deltas
MDNEXT_cci_pdf = d_cpdf(param.BinSize, 0, 1);
MDFEXT_cci_pdf = d_cpdf(param.BinSize, 0, 1);
% serially convolve FEXT/NEXT PDFs
for k=2:kxi
    if isequal(chdata(k).type, 'NEXT')
        MDNEXT_cci_pdf = conv_fct(MDNEXT_cci_pdf, chdata(k).pdfr);
    else % ... must be FEXT
        MDFEXT_cci_pdf = conv_fct(MDFEXT_cci_pdf, chdata(k).pdfr);
    end
end

% find "peaks" of MDNEXT/MDFEXT for reporting
mdnxi=find(cumsum(MDNEXT_cci_pdf.y)>=param.specBER, 1, 'first');
MDNEXT_peak_interference=abs(MDNEXT_cci_pdf.x(mdnxi));
mdfxi=find(cumsum(MDFEXT_cci_pdf.y)>=param.specBER, 1, 'first');
MDFEXT_peak_interference=abs(MDFEXT_cci_pdf.x(mdfxi));

% Combined crosstalk effect
cci_pdf = conv_fct(MDFEXT_cci_pdf, MDNEXT_cci_pdf);
cci_mxi=find(cumsum(cci_pdf.y)>=param.specBER, 1, 'first');
cci_msi=find(cumsum(cci_pdf.y)>=1e-12, 1, 'first');
cci_sigma=abs(cci_pdf.x(cci_msi)/(erfcinv(2*1e-12)*sqrt(2)));
crosstalk_max_peak_interference_at_BER=abs(cci_pdf.x(cci_mxi));

% combine cci and sci
isi_and_xtalk_pdf = conv_fct(sci_pdf, cci_pdf);
mxi=find(cumsum(isi_and_xtalk_pdf.y)>=param.specBER, 1, 'first');
max_peak_interference_at_BER=abs(isi_and_xtalk_pdf.x(mxi));


%% Equation 93A-42
combined_interference_and_noise_pdf = conv_fct(isi_and_xtalk_pdf, noise_pdf);

%% Equation 93A-34
combined_interference_and_noise_cdf=cumsum(combined_interference_and_noise_pdf.y);

%%  The noise amplitude, A_n, is the magnitude of the value of y0 that satisfies the relationship P(y0) = DER_0
A_n_ix=find(combined_interference_and_noise_cdf>param.specBER, 1, 'first');
A_n = abs(combined_interference_and_noise_pdf.x(A_n_ix));

%//    jpdfix=find(cumsum(isi_and_xtalk_pdf.y) >param.specBER,1,'first');

%% Equation 93A-1
COM=20*log10(max_signal/A_n);


if OP.RX_CALIBRATION ==1
    display ([' LOOP with WGN = ' num2str(param.WGN) ' performed with COM = ' num2str(COM) ])
    param.WGN=param.WGN +param.WGN_STEP;
    
    
end

%% reporting
output_args.channel_operating_margin_dB=COM;
if OP.RX_CALIBRATION== 1, output_args.WGN=param.WGN-param.WGN_STEP;end
output_args.peak_interference_mV=1000*A_n;
output_args.peak_channel_interference_mV=1000*max_peak_interference_at_BER;
%if exist('thru_max_peak_interference_at_BER','var'),
output_args.peak_ISI_mV=1000*thru_max_peak_interference_at_BER;
%end
%if exist('crosstalk_max_peak_interference_at_BER','var'),
output_args.peak_MDXTK_interference_mV=1000*crosstalk_max_peak_interference_at_BER;
%end
if ICN ~= 0,
    output_args.icn_mV=ICN*1000;
end
%if exist('MDNEXT_peak_interference','var'),
output_args.peak_MDNEXT_interference_mV=1000*MDNEXT_peak_interference;
%end
%if exist('MDFEXT_peak_interference','var'),
output_args.peak_MDFEXT_interference_mV=1000*MDFEXT_peak_interference;
%end
%if exist('max_signal','var'),
output_args.available_signal_after_eq_mV=1000*max_signal;

if OP.GET_FD ~=0 ,
    output_args.fit_loss_dB_at_Fnq = fit_loss;
    output_args.IL_dB_at_Fnq=Nq_loss;
    output_args.baud_rate_GHz=param.fb/1e9;
    output_args.ILD_RMS=ILD_RMS;
end
%end
%%
if OP.DEBUG
    cursors = d_cpdf(isi_and_xtalk_pdf.BinSize,param.AG*max_signal*[-1 1], [1 1]/2);
    signal_and_channel_noise_pdf = conv_fct(cursors, isi_and_xtalk_pdf);
    signal_and_system_noise_pdf = conv_fct(cursors, noise_pdf);
    signal_and_total_noise_pdf = conv_fct(cursors, combined_interference_and_noise_pdf);
    figure(600)
    semilogy(signal_and_channel_noise_pdf.x, abs(cumsum(signal_and_channel_noise_pdf.y)-0.5) ,'r','Disp','ISI and crosstalk')
    hold on
    semilogy(signal_and_system_noise_pdf.x, abs(cumsum(signal_and_system_noise_pdf.y)-0.5) ,'c','Disp','Jitter and system noise')
    semilogy(signal_and_total_noise_pdf.x, abs(cumsum(signal_and_total_noise_pdf.y)-0.5) ,'k','Disp','total noise PDF')
    
    ylim([param.specBER/10 1])
    ylabel('Probability')
    xlabel('volts')
    title(regexprep([chdata(1).base,' Voltage bathtub curves'],'_',' '))
    legend show
    
    display(['max noise at BER = ' num2str(max_peak_interference_at_BER)])
    display(['max signal after eq = ' num2str(param.AG*max_signal)])
end
% create mat file with result
if OP.DISPLAY_WINDOW
    if output_args.channel_operating_margin_dB >= param.cc1
        h = msgbox(['PASS ... margin = ' num2str(output_args.channel_operating_margin_dB) 'dB']); set(h,'Color','g');
    else
        h = msgbox(['FAIL ... margin = ' num2str(output_args.channel_operating_margin_dB) 'dB']); set(h,'Color','r');
    end
else
    if output_args.channel_operating_margin_dB >= param.cc1
        fprintf(['<strong>' 'PASS ... margin = ' num2str(output_args.channel_operating_margin_dB) 'dB' '</strong> ' '\n' ])
    else
        fprintf(2,['FAIL ... margin = ' num2str(output_args.channel_operating_margin_dB) 'dB\n']);
    end
    
end

if OP.DEBUG
    
    if OP.GET_FD ~=0, output_args.equivalent_ISI_ICN=sci_sigma; end
    output_args.ctle_zero_poles_acdcgaindB=[-result.eq.ctle.zero/(2*pi) -result.eq.ctle.pole2/(2*pi) -result.eq.ctle.pole1/(2*pi)];
    output_args.acdcgaindB=20*log10(result.eq.ctle.maxkacdc);
    output_args.txle_taps=result.eq.txle;
    output_args.dfe_taps=result.DFE_taps;
    if OP.GET_FD ~=0, output_args.sci_noise_FD_RMS=FD_rms_sci;end
    if xtk~=0
        output_args.cci_noise_TD_BER=crosstalk_max_peak_interference_at_BER;
        output_args.equivalent_ISI_ICN=cci_sigma;
    end
    output_args.max_peak_interference_at_BER=max_peak_interference_at_BER;
    output_args.FOM =result.best_FOM;
end

%% making mat file
output_args.dfe4_rss=norm(result.DFE_taps(4:end));
if OP.EXTERNAL== false,
    output_args.file_names=str2mat(chdata.base);
    display( ['Files set is: ' chdata(1).base]);
    if(OP.DEBUG), save ([OP.RESULT_DIR chdata(1).base '_results.mat'],'output_args');end
else
    output_args.baud_rate_GHz=param.fb/1e9;
    output_args.sigma_r=param.Na_rms;
    output_args.sigma_rj=param.sigma_RJ;
    output_args.Add_noise=param.A_DD;
end

%% making csv file
if OP.CSV_REPORT ==1
    ihead=1;
    header(ihead,:)={'File set'};
    data(ihead)={chdata(1).base};
    ihead=ihead+1;
    if exist('max_signal','var') ,
        header(ihead,:)={'COM_dB'};
        data(1,ihead)={output_args.channel_operating_margin_dB};
        ihead=ihead+1;
    end
    header(ihead,:)={'peak_interference_mV'};
    data(ihead)={output_args.peak_interference_mV};
    ihead=ihead+1;
    header(ihead,:)={'peak_channel_interference_mV'};
    data(ihead)={output_args.peak_channel_interference_mV};
    ihead=ihead+1;
    if exist('thru_max_peak_interference_at_BER','var'),
        header(ihead,:)={'peak_ISI_mV'};
        data(1,ihead)={output_args.peak_ISI_mV};
        ihead=ihead+1;
    end
    if exist('crosstalk_max_peak_interference_at_BER','var'),
        header(ihead,:)={'peak_MDXTK_interference_mV'};
        data(1,ihead)={output_args.peak_MDXTK_interference_mV};
        ihead=ihead+1;
    end
    if ICN ~= 0,
        header(ihead,:)={'ICN'};
        data(1,ihead)={output_args.icn_mV};
        ihead=ihead+1;
    end
    if exist('MDNEXT_peak_interference','var'),
        header(ihead,:)={'peak_MDNEXT_interference_mV'};
        data(1,ihead)={output_args.peak_MDNEXT_interference_mV};
        ihead=ihead+1;
    else
        header(ihead,:)={'peak_MDNEXT_interference_mV'};
        data(1,ihead)={0};
        ihead=ihead+1;
    end
    if exist('MDFEXT_peak_interference','var'),
        header(ihead,:)={'peak_MDFEXT_interference_mV'};
        data(1,ihead)={output_args.peak_MDFEXT_interference_mV};
        ihead=ihead+1;
    else
        header(ihead,:)={'peak_MDFEXT_interference_mV'};
        data(1,ihead)={0};
        ihead=ihead+1;
    end
    if exist('max_signal','var') ,
        if OP.GET_FD ~=0 ,
            header(ihead,:)={'fit_loss_dB_at_Fnq'};
            data(1,ihead)={output_args.fit_loss_dB_at_Fnq};
            ihead=ihead+1;
            header(ihead,:)={'IL_loss_dB_at_Fnq'};
            data(1,ihead)={output_args.IL_dB_at_Fnq};
            ihead=ihead+1;
            header(ihead,:)={'ILD RMS'};
            data(1,ihead)={output_args.ILD_RMS};
            ihead=ihead+1;
        end
        header(ihead,:)={'available_signal_after_eq_mV'};
        data(1,ihead)={output_args.available_signal_after_eq_mV};
        ihead=ihead+1;
    end
    header(ihead,:)={'dfe4 rss'};
    data(ihead)={output_args.dfe4_rss};
    ihead=ihead+1;
    header(ihead,:)={'coding'};
    ihead=ihead+1;
    header(ihead,:)={'Fnq (GHz)'};
    data(1,ihead)={1/param.ui/2};
    %//ihead=ihead+1;
    header_string=header(1);
    for i = 2:length(header)
        header_string = strcat(header_string, ',' ,header(i));
    end
    if OP.EXTERNAL== false
        fid = fopen([OP.RESULT_DIR chdata(1).base '_results.csv'],'w');
        fprintf(fid,'%s\r\n',cell2mat(header_string));
        fprintf(fid,'%s,',cell2mat(data(1)));
        for i=2:length(header)-2
            fprintf(fid,'%f,',cell2mat(data(i)));
        end
        fprintf(fid,'%s,',cell2mat({param.coding}));
        fprintf(fid,'%f,',cell2mat({1/param.ui/2/1e9}));
        for i=2:kxi
            fprintf(fid,'%s,',cell2mat({chdata(i).base}));
        end
        
        fclose(fid);
    end
end

%%
%--------------------------------------------------------------------------
%--------------------------------------------------------------------------
%--------------------------------------------------------------------------
%--------------------------------------------------------------------------
%--------------------------------------------------------------------------
%--------------------------------------------------------------------------
function [voltage, t_base] =s21_to_impulse_DC(IL, freq_array, time_step)
% 3/23/2012
% Creates a time-domain impulse response from frequency-domain IL data.
% IL does not need to have DC but a corresponding frequency array
% (freq_array) is required.
%
% Causality is imposed using the Alternating Projections Method. See also:
% Quatieri and Oppenheim, "Iterative Techniques for Minimum Phase Signal
% Reconstruction from Phase or Magnitude", IEEE Trans. ASSP-29, December
% 1981 (http://ieeexplore.ieee.org/xpls/abs_all.jsp?arnumber=1163714)

rel_tol = 1e-3;
diff_tol = 1e-4;
causality_tol = 0.05;
% spline fit to DC , rim 3-23-2012
ILin=IL;
fmax=1/time_step/2;
freq_step=(freq_array(3)-freq_array(2))/1;
fout=0:1/round(fmax/freq_step)*fmax:fmax;
IL=interp_Sparam(ILin,freq_array,fout);
for i=1:length(IL), if isnan(IL(i)), IL(i)=IL(i-1); else IL(i)=IL(i);end; end
IL = IL(:);
% add padding for time steps
IL_symmetric = [IL(1:end-1);0; flipud(conj(IL(2:end-1)))];
fir = real(ifft(IL_symmetric));
% Correct non-causal effects frequently caused by extrapolation of IL
abs_fir=abs(fir);
a = find(abs_fir>max(abs_fir)*causality_tol);
start_ind = a(1);
err=inf;
% figure(1000)
t_base = (0:length(fir)-1)/(freq_step*length(fir));
% UI_ones(1:32)=1;
% pulse=filter(UI_ones,1,fir);
% figure
% plot(t_base,pulse);
% hold on

enforce_causality=1;
while enforce_causality
    fir(floor(length(fir)/2):end)=0;
    fir(1:start_ind)=0;
    IL_modified=abs(IL_symmetric).*exp(1j*angle(fft(fir)));
    fir_modified = real(ifft(IL_modified));
    delta = abs(fir-fir_modified);
    
    err_prev = err;
    err=max(delta)/max(fir);
    if (err<rel_tol) || (err_prev-err<diff_tol)
        break;
    end
    
    fir=fir_modified;
end
% truncate final samples smaller than 1e-3 of the peak
fir_peak = max(abs(fir));
fir_last  = find(abs(fir)>fir_peak*1e-3, 1, 'last');

voltage = fir(1:fir_last);
t_base = t_base(1:fir_last);


function [Sout] = interp_Sparam(Sin,fin,fout)
% Sout = interp_Sparam(Sin,fin,fout)
%
% Interpolate S-parameters Sin from frequency grid fin to frequency grid
% fout.
% fout is assumed to begin at 0 (DC), and zero phase is enforced (see
% source for options).
% Extrapolation to higher frequencies is done using the last available
% frequency data (without fitting the trend). This is not exactly
% physical, but typically happens only at very high frequencies where
% IL is low - so for IL the error is small. RL at high frequencies may not
% be reliably extrapolated.

if ( fin(end)<fout(end) )
    %    warning('Channel high frequencies extrapolation might be inaccurate!');
end

H_mag = abs(Sin);
H_mag(H_mag<eps)=eps; % handle ill cases...
H_ph = unwrap(angle(Sin));
% For long delay channels, the result can turn anti-causal if frequency step is too coarse. Don't let the
% user ignore that.
if mean(diff(H_ph))>0
    error('Anti-causal response found. Finer frequency step is required for this channel');
end

opt_interp_Sparam_mag='linear_trend_to_DC';
switch opt_interp_Sparam_mag
    case 'linear_trend_to_DC'
        fin_x=fin;
        H_mag_x=H_mag(:);
        if fin(1)>0
            p=polyfit(fin(1:10), H_mag(1:10), 1);
            dc_trend_val=polyval(p, 0);
            fin_x=[0, fin_x];
            H_mag_x = [dc_trend_val; H_mag_x];
        end
        if fin(end)<fout(end)
            warn_state=warning('off', 'MATLAB:polyfit:RepeatedPointsOrRescale');
            mid_freq_ind=round(length(fin)/2);
            p=polyfit(fin(mid_freq_ind:end), H_mag(mid_freq_ind:end), 1);
            warning(warn_state);
            hf_trend_val=polyval(p, fout(end));
            if hf_trend_val>H_mag(end)
                hf_trend_val=H_mag(end);
            elseif hf_trend_val<eps
                hf_trend_val=eps;
            end
            fin_x=[fin_x, fout(end)];
            H_mag_x = [H_mag_x; hf_trend_val];
        end
        H_mag_i = interp1(fin_x, H_mag_x, fout, 'linear', 0);
    case 'trend_to_DC'
        % extrapolate to trend value at DC.
        fin_x=fin;
        H_mag_x=H_mag;
        if fin(1)>0
            p=polyfit(fin(1:10)', log10(H_mag(1:10)), 1);
            dc_trend_val=10^polyval(p, 0);
            fin_x=[0, fin_x];
            H_mag_x = [dc_trend_val; H_mag_x];
        end
        if fin(end)<fout(end)
            warn_state=warning('off', 'MATLAB:polyfit:RepeatedPointsOrRescale');
            mid_freq_ind=round(length(fin)/2);
            p=polyfit(fin(mid_freq_ind:end)', log10(H_mag(mid_freq_ind:end)), 1);
            warning(warn_state);
            hf_trend_val=10^polyval(p, fout(end));
            if hf_trend_val>H_mag(end)
                hf_trend_val=H_mag(end);
            end
            fin_x=[fin_x, fout(end)];
            H_mag_x = [H_mag_x; hf_trend_val];
        end
        H_mag_i = 10.^interp1(fin_x,log10(H_mag_x),fout,'linear', 'extrap');
    case 'extrap_to_DC_or_zero'
        % same as extrap_to_DC but detect AC-coupled channels and
        % extrapolate them to 0.
        if fin(1)>0 && 20*log10(H_mag(1))<-20
            % assume AC coupling, with 0 at DC
            H_mag_i = 10.^interp1([0, fin],[-100; log10(H_mag)],fout(fout<=fin(end)),'linear', 'extrap');
        else
            H_mag_i = 10.^interp1(fin, log10(H_mag), fout(fout<=fin(end)),'linear', 'extrap');
        end
        H_mag_i(fout>fin(end)) = H_mag(end);
    case 'extrap_to_DC'
        % first extrapolate down to DC, then use highest available frequency
        % for higher frequencies
        H_mag_i = 10.^interp1(fin,log10(H_mag),fout(fout<=fin(end)),'linear', 'extrap');
        H_mag_i(fout>fin(end)) = H_mag(end);
    case 'old'
        H_mag_i = interp1(fin,H_mag,fout,'linear','extrap');
    otherwise
        warning('opt_interp_Sparam_mag valid values are "old", "extrap_to_DC"');
end

H_ph_i = interp1(fin,H_ph,fout,'linear', 0);

opt_interp_Sparam_phase='trend_and_shift_to_DC';
switch opt_interp_Sparam_phase
    case 'old'
        H_ph_i = H_ph_i-H_ph_i(1);
    case 'zero_DC'
        H_ph_i(1) = 0;
    case 'interp_to_DC'
        if fin(1) ~= 0
            H_ph_i = interp1([0; fin(:)], [0; H_ph(:)], fout, 'linear', 'extrap');
        end
    case 'interp_and_shift_to_DC'
        if fin(1) ~= 0
            dc_phase_trend = H_ph(1)-(H_ph(2)-H_ph(1))/(fin(2)-fin(1))*fin(1);
            H_ph_i = interp1([0; fin(:)], [0; H_ph(:)-dc_phase_trend], fout, 'linear', 'extrap');
        end
    case 'trend_and_shift_to_DC'
        % estimate low frequency group delay
        group_delay = -diff(H_ph(:))./diff(fin(:));
        low_freq_gd = group_delay(1:50);
        %  calculate trend, throwing away outliers
        m = median(low_freq_gd); sigma = std(low_freq_gd);
        lf_trend = mean(low_freq_gd(abs(low_freq_gd-m)<sigma));
        % correct outliers in first 10 phase samples
        for k=10:-1:1
            H_ph(k) = H_ph(k+1) + lf_trend*(fin(k+1)-fin(k));
        end
        fin_x=fin;
        H_ph_x=H_ph(:);
        % shift all phase data so that DC extrapolation to 0 follows trend
        if fin(1) ~= 0
            dc_phase_trend = H_ph(1)+lf_trend*(fin(1)-0);
            fin_x=[0, fin_x];
            H_ph_x=[0; H_ph(:)-dc_phase_trend];
        end
        % Modification: extrapolate using trend. (interp1 with "extrap" extrapolates using just
        % the last two samples, so noise can create an inverted slope and
        % non-causal response).
        if fout(end)>fin(end)
            group_delay = -diff(H_ph_x(:))./diff(fin_x(:));
            % p=polyfit(fin_x', H_ph_x, 1);
            hf_phase_trend = H_ph_x(end)-median(group_delay)*(max(fout)-max(fin_x));
            % hf_phase_trend=polyval(p,max(fout));
            fin_x=[fin_x, fout(end)];
            H_ph_x=[H_ph_x; hf_phase_trend];
        end
        H_ph_i = interp1(fin_x, H_ph_x, fout, 'linear', 'extrap');
        
    otherwise
        warning('debug_interp_Sparam valid values are "old", "zero_DC", "interp_to_DC", "interp_and_shift_to_DC"');
end
H_i = H_mag_i.*exp(1j*H_ph_i);
Sout=H_i;

function [data, SDD, SDC] = read_p4_s4params(infile, plot_ini_s_params, plot_dif_s_params, ports,OP)
%% FUNCTION :: read_sp4_sparams
%
% Description
%   Read the fid of single-ended 4-port complex S-parameters
%   in Touchstone format 'file' and convert to the internal
%   format using the port transform 'ports'
%
%   Created by Mike Y. He
%   April 22, 2005
%
%   Reused some code from
%   Anthony Sanders, Alex Deas, Bob Davidov (24 January 2005)
%   for touchstone 4-port S-matrix import.
%
%   Modified (2012-July-27) by Ken Young to match current indexing scheme and
%   optimized for quicker parameter matching and parsing. also, separated out
%   the plotting algorithms into their own sub-function routines
%
% Input Variables (required)
%   file                -- The s4p file to be read and cover ted
%   plot_ini_sparams    -- Plot the initial s-parameter information. For debugging purposes
%   plot_diff_sparams   -- Plot the differential s-parameter information. For debugging purposes
%   swap                -- Re-order the port layout
%
% Output/Return Variables
%   sch         -- the data matrix contain the network parameter data points
%   schFreq     -- the frequency vector of the network parameter data points
%   sdc         -- the differential in/common-mode out s-parameter data matrix
%   sdd         -- the differential in/differential out s-parameter data matrix
%
% Local Variables (scoped locally to current function only)
%   fid                 -- the opened file being processed
%   fileDataLine        -- the current line in the fid
%   optionLine          -- the options line in the s4p file
%   ports               -- the order of the ports being processed
%   idxNetParams        -- the position (line) of the network parameter data points in the s4p file
%   netParamDataPoints  -- the network parameter data points of the current line of the fid
%   schFreq             -- the frequency points in the s4p file stored as a separate vector
%   sch                 -- the network parameter data points from the s4p file converted in matrix form
%   S                   --
%   T                   --
%   W                   --
%   D                   --
%   sdd                 -- the differential in/differential out form of network parameter data points
%   sdc                 -- the differential in/common-mode out form of network parameter data points
%   plot_ini_sparams    -- Plot the initial s-parameter information. For debugging purposes
%   plot_diff_sparams   -- Plot the differential s-parameter information. For debugging purposes
%

% open the fid
fid = fopen(infile, 'r');

data_file = textscan(fid, '%s', 'delimiter', sprintf('\n'));
data_file = data_file{1,1}; % removes a dimension layer of cell array depth

fclose(fid);

% backwards compatibility settings. can be removed in updated code.
if ~exist('OP', 'var'); OP.DISPLAY_WINDOW = true; end
if isempty(ports); ports = [1 3 2 4]; end % default order normally used.

if OP.DISPLAY_WINDOW
    set(0,'defaulttextinterpreter','none') % prevents subscripting character in displayed messages
    hMsgBox = msgbox(infile, 'Reading S-Parameter File'); % display a progress bar for reading the s-parameter file(s)
end

% parse the file array until the s-parameters options line is found
idxLine = 1;
while true
    if regexp(data_file{idxLine}, '^#', 'once')
        % tests the option fileDataLine cell array to see if it contains...
        if ~isempty(regexpi(data_file{idxLine}, '\sGHz')); freq_scale = 1e9; end
        if ~isempty(regexpi(data_file{idxLine}, '\sMHz')); freq_scale = 1e6; end
        if ~isempty(regexpi(data_file{idxLine}, '\sKHz')); freq_scale = 1e3; end
        if ~isempty(regexpi(data_file{idxLine}, '\sHz'));  freq_scale = 1e0; end
        if ~isempty(regexpi(data_file{idxLine}, 'RI')) && ~isempty(regexpi(data_file{idxLine}, 'R'))
            format = 'complex';
        end
        if ~isempty(regexpi(data_file{idxLine}, 'MA')) && ~isempty(regexpi(data_file{idxLine}, 'R'))
            format = 'linpolar';
        end
        if ~isempty(regexpi(data_file{idxLine}, 'DB')) && ~isempty(regexpi(data_file{idxLine}, 'R'))
            format = 'dBpolar';
        end
        
        % if the format or freq_scale are not set then exit with error
        if isempty(freq_scale)
            error( 'read_s4p:MissingOptionsParameter', ['\n\t THE FREQUENCY SCALE PARAMETER INFORMATION IS MISSING' ...
                'FROM THE OPTIONS LINE IN THE S4P FILE']);
        elseif isempty(format)
            error( 'read_s4p:MissingOptionsParameter', ['\n\t THE FORMAT PARAMETER INFORMATION IS MISSING' ...
                'FROM THE OPTIONS LINE IN THE S4P FILE']);
        end
        
        idxLine = idxLine+1; %step one more line before exiting loop
        break
    end
    idxLine = idxLine+1;
end

col_1=1; col_2=2; col_3=3; col_4=4; % This is to maintain current indexing compatibility as well as readability
schFreqAxis = zeros(1,int8((length(data_file)-idxLine)/4)); % preallocated array length for memory and time purposes
idxNetParams = 1; % each 4-port network parameter data point set will receive an index

freqCounter = 0; % variable that will be used to capture number of freq points.

while true
    if isempty(data_file{idxLine}) || ~isempty(regexp(data_file{idxLine}, '^!', 'once'))
        idxLine = idxLine+1;
    else
        for row=1:4
            tmp = textscan(data_file{idxLine}, '%f' , 'MultipleDelimsAsOne', ' ');
            dataPointsLine = tmp{1,1}; % removes a dimension layer of cell array depth
            if mod(length(dataPointsLine),8)==1
                schFreqAxis(idxNetParams) = dataPointsLine(1) * freq_scale;
                dataPointsLine(1) = []; % removes the frequency value from the current data fileDataLine cell array
                % so that only the network parameter data is in the array
            end
            if mod(length(dataPointsLine),8)~=0  % sanity check to make sure the network parameter data point line structure is correct
                msg = sprintf(['\n'...
                    '\n\t ******************************************************' ...
                    '\n\t ** There is an error in the s4p network parameter   **' ...
                    '\n\t ** data point information structure at line %G.'        ...
                    '\n\t **              ENDING PROGRAM                      **' ...
                    '\n\t ******************************************************' ...
                    '\n\n\n\t Press any key to exit (an error will be displayed). \n\n']);
                disp(msg)
                pause on;
                pause;
                pause off;
                error( 'warning:SparamFileDataError', '\n\t INCORRECT S4P NETWORK PARAMETER DATA POINT STRUCTURE');
            else
                switch format
                    case 'complex'
                        sch(idxNetParams, ports(row), ports(col_1)) = dataPointsLine(1) + 1i*dataPointsLine(2);
                        sch(idxNetParams, ports(row), ports(col_2)) = dataPointsLine(3) + 1i*dataPointsLine(4);
                        sch(idxNetParams, ports(row), ports(col_3)) = dataPointsLine(5) + 1i*dataPointsLine(6);
                        sch(idxNetParams, ports(row), ports(col_4)) = dataPointsLine(7) + 1i*dataPointsLine(8);
                    case 'linpolar'
                        sch(idxNetParams, ports(row), ports(col_1)) = dataPointsLine(1) * exp(1i*dataPointsLine(2)*(pi/180));
                        sch(idxNetParams, ports(row), ports(col_2)) = dataPointsLine(3) * exp(1i*dataPointsLine(4)*(pi/180));
                        sch(idxNetParams, ports(row), ports(col_3)) = dataPointsLine(5) * exp(1i*dataPointsLine(6)*(pi/180));
                        sch(idxNetParams, ports(row), ports(col_4)) = dataPointsLine(7) * exp(1i*dataPointsLine(8)*(pi/180));
                    case 'dBpolar'
                        sch(idxNetParams, ports(row), ports(col_1)) = 10^(dataPointsLine(1)/20) * exp(1i*dataPointsLine(2)*(pi/180));
                        sch(idxNetParams, ports(row), ports(col_2)) = 10^(dataPointsLine(3)/20) * exp(1i*dataPointsLine(4)*(pi/180));
                        sch(idxNetParams, ports(row), ports(col_3)) = 10^(dataPointsLine(5)/20) * exp(1i*dataPointsLine(6)*(pi/180));
                        sch(idxNetParams, ports(row), ports(col_4)) = 10^(dataPointsLine(7)/20) * exp(1i*dataPointsLine(8)*(pi/180));
                    otherwise
                end
            end
            idxLine = idxLine+1;
        end
        
        freqCounter = freqCounter+1; % capturing number of freq points for preallocation.
        idxNetParams = idxNetParams + 1;
    end
    if idxLine > length(data_file)
        break
    end
end

% calculate differential s parameter matrix from single ended
for i=1:size(sch,1)
    S(:,:) = sch(i,:,:);
    T = [1 1 0 0 ; 1 -1 0 0 ; 0 0 1 1 ; 0 0 1 -1];
    W = T * (S / T);
    D(i,:,:) = W(:,:);
end

% D matrix should be
% Scc11 Scd11 Scc12 Scd21
% Sdc11 Sdd11 Sdc12 Sdd12
% Scc21 Scd21 Scc22 Scd22
% Sdc21 Sdd21 Sdc22 Sdd22

% proper values
sdd(:,1,1) = D(:,2,2);
sdd(:,2,2) = D(:,4,4);
sdd(:,1,2) = D(:,2,4);
sdd(:,2,1) = D(:,4,2);

sdc(:,1,1) = D(:,2,1);
sdc(:,2,2) = D(:,4,3);
sdc(:,1,2) = D(:,2,3);
sdc(:,2,1) = D(:,4,1);


if (plot_ini_s_params == 1)
    figure
    for mj=1:4
        legendIndex=1;
        for mi=1:4
            subplot(2,2,mj);
            plot(data.f, 20*log10(abs(data.m(:,mj,mi)+1.0e-15)), c(mj,mi), 'linewidth',2); hold on
            legendTitle(legendIndex,1:3) = sprintf('S%d%d', mj, mi);
            legendIndex=legendIndex+1;
        end
        xlabel('Frequency (Hz)');
        ylabel('Magnitude (dB)');
        legend(legendTitle,3);
        grid on
        title(infile);
    end
end

if (plot_dif_s_params == 1)
    figure
    subplot(2,1,1);
    legendIndex=1;
    for mj=1:2
        for mi=1:2
            plot(data.f, 20*log10(abs(SDD(:,mj,mi))), c(1, (mj-1)*2+mi), 'linewidth',2); hold on
            legendTitle(legendIndex, 1:5) = sprintf('SDD%d%d', mj, mi);
            legendIndex=legendIndex+1;
        end
    end
    xlabel('Frequency (Hz)');
    ylabel('Magnitude (dB)');
    legend(legendTitle,3);
    grid on
    title(infile);
    
    subplot(2,1,2);
    legendIndex=1;
    for mj=1:2
        for mi=1:2
            plot(data.f, 20*log10(abs(SDC(:,mj,mi))+1.0e-15), c(1, (mj-1)*2+mi), 'linewidth',2); hold on
            legendTitle(legendIndex,1:5) = sprintf('SDC%d%d', mj, mi);
            legendIndex=legendIndex+1;
        end
    end
    xlabel('Frequency (Hz)');
    ylabel('Magnitude (dB)');
    legend(legendTitle,3);
    grid on
    
end


if OP.DISPLAY_WINDOW; close(hMsgBox); end
% backwards compatibility output variables
data.m = sch;
schFreqAxis=schFreqAxis(1:freqCounter); % truncating preallocated array to number of freq points.
data.freq = schFreqAxis;
SDC = sdc;
SDD = sdd;
% end read_sp4_sparam
function result = readdataSnPx(filename, nport)
%function [freq, cs] = readdataSnPx(filename, nport)
% [freq, cs] = readdataSnP(filename, nport, format, nheader)
%
% Read Touchstone file with frequencies in units of Hertz
%
% Input:
% ======
% filename: Name of the Touchstone/SnP file
% nport: Number of ports
% format: 'RI' for real/imag, 'MA' for mag/angle (check option line in the
%         Touchstone file)
% nheader: Number of header lines (comment lines plus option line in the
%          Touchstone file)
%
% Output:
% =======
% freq: Vector of frequencies [Hz]
% cs: 3-D array of complex-valued S parameters where cs(i,j,k) is S(i,j)
%     at frequency freq(k)
%
% Note: If frequency unit is not Hertz (but GHz, MHz etc.) simply scale
% frequencies appropriately after reading the data.
%
% Ref.: Touchstone(R) File Format Specification, Rev.1.1,
% EIA/IBIS Open Forum, 2002.
%
% Written by Henning Braunisch, September 2004.
% Updated by Steven Krooswyk, April 2006.


fid = fopen(filename, 'r');


% Skip header lines
str = ' ';
n = 0;
while ~strcmp(str(1),'#')
    str = fgetl(fid);
    if length(str) == 0
        str=' ' ;
        if n > 1000
            display('error: could not find config line (#)')
            break
        end
    end
    n = n + 1;
end

% parse configuration line
A=sscanf(str,'%1s %2s %1s %2s %1s %2s',[1,inf]);
p = find(A=='S');           %position of 'S'
units = lower(A(2:p-1));    %units before 'S'
format = A(p+1:p+2);        %format after 'S'

% skip any more header lines
%while ~str

nk = 0; % frequency counter
while 1==1
    
    [temp, count] = fscanf(fid, '%f', 1);
    if count == 0
        [temp2, count2] = fscanf(fid, '%s', 1);
        if length(temp2)>0, fgetl(fid); continue, end;
        break
    end
    nk = nk+1; freq(1,nk) = temp;
    for ni = 1:nport
        for nj = 1:nport
            if lower(format) == 'ma'
                mag = fscanf(fid, '%f', 1); ang = fscanf(fid, '%f', 1);
                cs(ni,nj,nk) = mag * exp(1i*ang*pi/180);
            elseif lower(format) == 'ri'
                re = fscanf(fid, '%f', 1); im = fscanf(fid, '%f', 1);
                cs(ni,nj,nk) = re + 1i*im;
            elseif lower(format) == 'db'
                db = fscanf(fid, '%f', 1); ang = fscanf(fid, '%f', 1);
                M = 10^(db/20);
                %re = M*cos(ang);
                %im = M*sin(ang);
                re =  M*cos(ang * pi / 180);
                im =  M*sin(ang * pi / 180);
                cs(ni,nj,nk) = re + 1i*im;
            else
                error('readdataSnP: Unknown data format');
            end
        end
    end
end

fclose(fid);

% If 2-port then swap S_12 and S_21 per Touchstone spec
if nport == 2
    temp = cs(2,1,:);
    cs(2,1,:) = cs(1,2,:);
    cs(1,2,:) = temp;
end

% Update freq units to Hz
switch lower(units)
    case 'hz'
        
    case 'khz'
        freq=freq.*1e3;
    case 'mhz'
        freq=freq.*1e6;
    case 'ghz'
        freq=freq.*1e9;
end

% passivity check


result.freq = freq;
result.cs    = cs;

function result = reduce(var1)
% --- Reduce 1x1xn array to 1xn
out = zeros(1,length(var1));
out(1,:) = var1(1,1,:);
result=out;
% end % reduce

function result=get_SBR(ir,samp_UI)
%ir = impulse response
%t_base=time array with equal time steps
%samp_UI = number of samples per UI for ir
%   result.SBR
UI_ones(1:samp_UI)=1;
%pulse=conv(ir,UI_ones);
pulse=filter(UI_ones,1,ir);
result.SBR=pulse;


function [pdf]=get_pdf(max_signal, excluded_UI, i, chdata, param, OP)
index=i;
SBR=chdata(i).result.SBR(:)'; % row vector
type=chdata(i).type;
samp_UI=param.samples_per_ui;
name=chdata(i).base;
if param.levels == 2 ; coding='NRZ' ; else coding='PAM4'; end
residual_response = SBR;
if isequal(type, 'THRU')
    % for thru pulse response:
    % remove the cursor and the DFE postcursors (up to their limit), since
    % we only care about the residuals.
    ideal_cancelled_cursors = SBR(param.tc+param.samples_per_ui*(0:param.ndfe));
    effective_cancelled_cursors = sign(ideal_cancelled_cursors) .* ...
        min(abs(ideal_cancelled_cursors), residual_response(param.tc)*[1, param.bmax]);
    effective_cancellation_samples = kron(effective_cancelled_cursors, ones(1, param.samples_per_ui));
    
    % Apply a constant DFE coefficient 1/2 UI before and after each postcursor. Not
    % really needed for COM, but helps debugging. May be factored out in future revisions.
    start_cancel = param.tc-param.samples_per_ui/2;
    end_cancel = param.tc+(1/2+param.ndfe)*param.samples_per_ui - 1;
    residual_response(start_cancel:end_cancel) = ...
        residual_response(start_cancel:end_cancel) - effective_cancellation_samples;
else
    % for crosstalk pulse responses, nothing is cancelled, and all phases
    % are equally important.
    param.tc=1;
end

nui=round(length(residual_response)/param.samples_per_ui);

for i=1:param.samples_per_ui
    vs(1:nui-2,i)=residual_response(param.samples_per_ui*(1:nui-2)+i);
end

if OP.DISPLAY_WINDOW,
    hwaitbar=waitbar(0);
else
    fprintf(' %i',index)
end

% determine which pdf to use
if isequal(type, 'THRU')
    % one phase is interesting for thru
    phases = mod(param.tc,param.samples_per_ui);
    if phases==0, phases = param.samples_per_ui; end
else
    phases=1:samp_UI;
end
for k=phases
    pdf_samples(k)=get_pdf_from_sampled_signal(vs(:,k), param.levels, param.BinSize);
    mxV(k)=sqrt(sum( pdf_samples(k).x.^2.*pdf_samples(k).y)); % standard deviation of PDF
    progress = k/length(phases);
    if OP.DISPLAY_WINDOW, waitbar(progress, hwaitbar, ['processing COM pdf ' num2str(index) ] ); drawnow; end
    if OP.DEBUG
        figure(index+150);
        subplot(2,1,2);
        semilogy(pdf_samples(k).x, pdf_samples(k).y,'disp',num2str(k) );
        hold on;
        title(strrep(name,'_',' '))
    end
end
[na pxi]=max(mxV);
pdf=pdf_samples(pxi);

if OP.DISPLAY_WINDOW,
    close(hwaitbar);
end


function [ILN, FIT]= get_ILN(sdd21,faxis_f2)
% used for FD IL fitting
% sdd21 us a complex insertion loss
fmbg=[ones(length(faxis_f2),1).*transpose(sdd21)  transpose(sqrt(faxis_f2)).*transpose(sdd21)  transpose(faxis_f2).*transpose(sdd21) transpose(faxis_f2.^2).*transpose(sdd21) ];
warning('off','MATLAB:nearlySingularMatrix');%display('ignore the following singularity warning that is expected')
unwraplog=log(abs(sdd21))+1i*unwrap(angle(sdd21));
LGw=transpose([sdd21.*unwraplog ]);
alpha = ((fmbg'*fmbg)^-1)*fmbg'*LGw;
efit=(alpha(1)+alpha(2).*sqrt(faxis_f2)+alpha(3).*faxis_f2 +faxis_f2.^2.*alpha(4)   );
FIT=transpose(exp(transpose(efit)));
ILN = sdd21-FIT;

function [result BC param chdata]=optimize_fom( OP, param, chdata) % TODO: call this FOM
%% input
% chdata(1).ir is the impulse response input expected to be normalized to At, peak drive voltage
% baud_rate - baud rate in seconds
% param.samples_per_ui = samples per UI of IR
% param.max_ctle - maximum ac to dc gain in dB
% param.tx_ffe(1) - maximum pre cursor (positive value)
% param.tx_ffe(2) - maximum post cursor (positive value)
% param.tx_ffe_step - sweep step size for tx pre and post taps
% param.ndfe - number of reference dfe taps
% output
% result.eq.txle - [ precusor curosr postcursor]: pre and post are negative
% result.eq.ctle.zero - Frequency of zero (HZ)
% result.eq.ctle.pole1- Frequency of pole (HZ)
% result.eq.ctle.pole2- -infinity
% result.eq.ctle.maxgain - real gain dB
% result.eq.ctle.maxkacdc - max acdc gain (not dB)
% result.IR - impulse response
% result.PR - pulse response (SBR)
% result.avail_signal - maximum signal after equalization
% result.avail_sig_index - index in result.IR of max signal
% result.best_FOM - best raw ISI
% BC return is for printing data
%if OP.DEBUG ~= 0 display('eq ver 1.2') ;end
min_number_of_UI_in_response=40;
baud_rate=1/param.ui;
%fb=baud_rate;
cm1_values = 0:param.tx_ffe_step:param.tx_ffe(1);
cp1_values = 0:param.tx_ffe_step:param.tx_ffe(2);
ctle_values= 0:param.ctle_step:param.max_ctle; % assume  param.max_ctle is a the DC dB loss i.e. use a positive number
kacdc_values = 10.^(-ctle_values/20); % and negate here
best_ctle = [];
best_FOM = -inf;
best_setting = [];
delta_sbr = [];
%t=(-param.samples_per_ui*5):(param.samples_per_ui*20);
%time_step=1/baud_rate/param.samples_per_ui;
pxi=0;
if ~OP.DISPLAY_WINDOW, fprintf('COM training '); end
for ctle_index=1:length(kacdc_values)
    kacdc=kacdc_values(ctle_index);
    best_sbr_h = [];
    if OP.DISPLAY_WINDOW, hwaitbar=waitbar(0);end
    for k=1:param.K
        [chdata(k).impulse_response p1_ctle, p2_ctle, z_ctle] = TD_CTLE(chdata(k).ir, baud_rate,kacdc,param.samples_per_ui);
    end
    if OP.RX_CALIBRATION
        [ sigma_noise sigma_N ] = get_sigma_noise( kacdc, param, chdata);
    else
        %% Equation 93A-20 %%
        H_ctle=(4i*chdata(1).faxis/param.fb + kacdc)./((1+1i*chdata(1).faxis/param.fb).*(1+4i*chdata(1).faxis/param.fb));
        %% Equation 93A-32 - independent of FFE setting%%
        sigma_N = sqrt(param.eta_0*sum(abs(param.H_r(2:end) .* H_ctle(2:end)).^2 .* diff(chdata(1).faxis)/1e9));
    end
    
    for k_cm1=1:length(cm1_values)
        cm1=cm1_values(k_cm1);
        for k_cp1=1:length(cp1_values)
            cp1=cp1_values(k_cp1);
            pxi=pxi+1;
            progress = pxi/( length(cm1_values)*length(cp1_values)*length(kacdc_values) );
            if OP.DISPLAY_WINDOW
                waitbar(progress, hwaitbar, 'Processing THRU SBR'); drawnow;
            else
                if ~mod(progress*100,20), fprintf([ '%i' '%% ' ], progress*100 );end
            end
            txffe = [-cm1, 1-cm1-cp1, -cp1];
            if txffe(2)<max(abs(txffe))
                continue;
            end
            upsampled_txffe = zeros(1, param.samples_per_ui*2+1); % start with zeros everywhere
            upsampled_txffe(1+(0:2)*param.samples_per_ui) = txffe; % plant the coefficients in the desired locations
            effective_channel = filter(upsampled_txffe, 1, chdata(1).impulse_response);
            sbr=filter(ones(param.samples_per_ui, 1), 1, effective_channel);
            [UNUSED_OUTPUT, sbr_peak_i]=max(abs(sbr));
            triple_transit_time = round(sbr_peak_i*2/param.samples_per_ui)+20;
            if min_number_of_UI_in_response < triple_transit_time
                min_number_of_UI_in_response = triple_transit_time;
            end
            
            %% initial guess at cursor location (t_s)  - based on approximate zero
            %% crossing
            zxi = find(diff(sign(sbr-.01*max(sbr)))>=1); % added the =
            zxi = zxi(zxi<sbr_peak_i);
            zxi = zxi(sbr_peak_i - zxi < 4*param.samples_per_ui);% made 4 UI to accommodate high dispersion channels
            if isempty(zxi)
                continue;
            elseif length(zxi)>1
                zxi=zxi(end);
            end
            cursor_i = zxi+param.samples_per_ui;
            
            if param.ndfe==0 % TODO: remove this?
                max_dfe1=0;
            else
                max_dfe1=1; % assuming bmax for h(1) is fixed at 1; should change if it is a parameter
            end
            %% adjust cursor_i to Solve equation 93A-23 %%
            % Mouller-Muller criterion with DFE
            mm_range = cursor_i+(0:2*param.samples_per_ui);
            mm_metric = ...
                abs(sbr(mm_range-param.samples_per_ui) - max(sbr(mm_range+param.samples_per_ui)-max_dfe1*sbr(mm_range), 0));
            [UNUSED_OUTPUT, mm_cursor_offset] = min(mm_metric);
            cursor_i = cursor_i+mm_cursor_offset-1;
            
            cursor = sbr(cursor_i);
            %% 93A.1.6 step c defines A_s %%
            A_s = cursor/(param.levels-1);
            if isempty(delta_sbr)
                delta_sbr = sbr;
                delta_cursor_i = cursor_i;
                delta_cursor = cursor;
            end
            sbr=sbr(:);
            param.sigma_xl=0; % .003; %planning improving finding FOM later
            if OP.DEBUG && OP.BREAD_CRUMBS
                BC(ctle_index,k_cm1,k_cp1).sbr=sbr; % hold on to sbr for debug
            else
                BC=0;
            end
            
            %% Equation 93A-25 "otherwise" case %%
            far_cursors = sbr(cursor_i+param.samples_per_ui*(param.ndfe+1):param.samples_per_ui:end);
            precursors = sbr(cursor_i+param.samples_per_ui*(-5:-1));
            
            %% Error message if the sbr is not long enough for the specified range of Nb %%
            if length(sbr) < cursor_i+param.samples_per_ui*(param.ndfe+1)
                error_msg = char(strcat('Pulse Response contains',{' '}, num2str(length(sbr)-cursor_i), {' '},'samples after the cursor.  Specified Nb requires',{' '}, num2str(param.samples_per_ui*(param.ndfe+1)), {' '},'samples after the cursor'));
                msgbox(error_msg, 'Error')
                close(hwaitbar);
                error('Pulse Response contains %s samples after the cursor. Specified Nb requires %s samples after the cursor',num2str(length(sbr)-cursor_i),num2str(param.samples_per_ui*(param.ndfe+1)));
            end
            
            %% Equation 93A-25, when 1<=n<=N_b
            dfecursors=sbr(cursor_i+param.samples_per_ui*(1):param.samples_per_ui:cursor_i+param.samples_per_ui*(param.ndfe));
            excess_dfe_cursors = dfecursors - ...
                sign(dfecursors) .*min(sbr(cursor_i)*param.bmax(:), abs(dfecursors) );
            
            %% Eq. 93A-26 %%
            sampling_offset = mod(cursor_i, param.samples_per_ui);
            %ensure we can take early sample
            if sampling_offset<=1
                sampling_offset=sampling_offset+param.samples_per_ui;
            end
            cursors_early_sample = sbr(sampling_offset-1:param.samples_per_ui:end);
            cursors_late_sample = sbr(sampling_offset+1:param.samples_per_ui:end);
            % ensure length are equal
            cursors_early_sample = cursors_early_sample(1:length(cursors_late_sample));
            h_J = (cursors_late_sample-cursors_early_sample)/2*param.samples_per_ui;
            
            %% Equation 93A-28 %%
            sigma_ISI = param.sigma_X*norm([precursors; excess_dfe_cursors; far_cursors]);
            
            %% Equation 93A-29 %%
            sigma_J = norm([param.A_DD param.sigma_RJ])*param.sigma_X*norm(h_J);
            
            %% Equations 93A-30 and 93A-31 %%
            sigma_XT =  get_xtlk_noise( kacdc, upsampled_txffe, param, chdata );
            
            %% Equation 93A-33 denominator (actually its sqrt)
            total_noise_rms = norm([sigma_ISI sigma_J sigma_XT sigma_N]);
            
            %% Equation 93A-33 (note log argument is voltage rather than power ratio)
            FOM = 20*log10(A_s/total_noise_rms);
            
            if (FOM > best_FOM) %&& (abs(sbr(cursor_i-param.samples_per_ui)) <= max(sbr)/20)
                egn=sbr((cursor_i+param.samples_per_ui*2):(cursor_i+param.samples_per_ui*param.ndfe));
                %                 if max(abs(egn))/sbr(cursor_i) >= param.bmax
                %                     continue;
                %                 end
                % need to set best if loop always continues ie to much ISI
                best_setting = txffe;
                best_sbr = sbr;
                best_ctle.kacdc =kacdc;
                best_ctle.p1=p1_ctle; % for a 1 pole does this need to be inf
                best_ctle.p2 = p2_ctle;
                best_ctle.z = z_ctle;
                best_FOM = FOM;
                best_cursor_i = cursor_i;
                best_IR=effective_channel;
                best_sigma_N = sigma_N;
                best_h_J = h_J;
                best_A_s=A_s;
                %           fprintf('found new best setting: %s\n', mat2str(txffe));
            end
        end
    end
end

if ~exist('best_cursor_i')% take last setting
    best_setting = txffe;
    best_sbr = sbr;
    best_ctle.kacdc =kacdc;
    best_ctle.p1=p1_ctle; % for a 1 pole does this need to be inf
    best_ctle.p2 = p2_ctle;
    best_ctle.z = z_ctle;
    best_FOM = FOM;
    best_cursor_i = cursor_i;
    best_IR=effective_channel;
    best_sigma_N = sigma_N;
    best_h_J = h_J;
end

best_cursor = best_sbr(best_cursor_i);
pre_emphasis_loss = 20*log10(best_cursor/delta_cursor);
% report during debug
PRin=filter(ones(param.samples_per_ui, 1),1, chdata(1).ir);
f=1e8:1e8:100e9;
%ctle_gain=(1+1i*2*pi*f/best_ctle.z)./((1+1i*2*pi*f/best_ctle.p1).*(1+1i*2*pi*f/best_ctle.p2))/best_ctle.kacdc;
ctle_gain=baud_rate*(1i.*f+0.25*baud_rate*best_ctle.kacdc)./((1i*f+0.25*baud_rate).*(1i*f+baud_rate));
lsbr=length(sbr);
t=0:param.ui/param.samples_per_ui:(lsbr-1)*param.ui/param.samples_per_ui;

sampled_best_sbr_precursors_t = (best_cursor_i/param.samples_per_ui+(-5:-1))*param.ui;
sampled_best_sbr_precursors = best_sbr(round(sampled_best_sbr_precursors_t/param.ui*param.samples_per_ui));
sampled_best_sbr_postcursors_t = (best_cursor_i:param.samples_per_ui:lsbr)/param.samples_per_ui*param.ui;
sampled_best_sbr_postcursors_t = sampled_best_sbr_postcursors_t(2:end); % exclude cursor
sampled_best_sbr_postcursors = best_sbr(round(sampled_best_sbr_postcursors_t/param.ui*param.samples_per_ui));
sampled_best_sbr_dfecursors_t = (best_cursor_i/param.samples_per_ui+(1:param.ndfe))*param.ui;
% apply max tap value constraint
dfe_cursors = sampled_best_sbr_postcursors(1:param.ndfe);
DFE_taps = dfe_cursors/best_cursor;

sampled_best_sbr_postcursors(1:param.ndfe)= dfe_cursors-sign(dfe_cursors).*min(best_cursor*param.bmax(:), abs(dfe_cursors) );

if OP.DEBUG ~=0
    display(['FOM:                ' ,num2str(best_FOM, 2),' dB']);
    display(['TXFFE coefficients: ' ,mat2str(best_setting) ] );
    display(['CTLE DC gain:       ' ,num2str(20*log10(best_ctle.kacdc)), ' dB']);
    display(['CTLE peaking gain:  ' ,num2str(20*log10(max(abs(ctle_gain))), 2), ' dB']);
    display(['Available signal:   ' ,num2str(best_cursor)]);
    figure(101)
    semilogx(f,20*log10(abs(ctle_gain)))
    hold on
    fbaud_tick=find(f >= baud_rate); fbaud_tick =fbaud_tick(1);
    fnq_tick=find(f >= baud_rate/2); fnq_tick =fnq_tick(1);
    stem(f(fnq_tick),20*log10(abs(ctle_gain(fnq_tick))),'g')
    stem(f(fbaud_tick),20*log10(abs(ctle_gain(fbaud_tick))),'g')
    title('CTLE response')
    ylabel('dB')
    xlabel('Hz')
    hold off
    figure(100)
    plot(t,best_sbr,'disp', [param.base 'SBR w Tx/Rx and pkg filters plus ffe and CTLE equalized']);
    hold on
    PRplt(1:param.samples_per_ui+5)=PRin(1); % line up with the ffe introduced delay
    PRplt(param.samples_per_ui+6:length(t))=PRin(1:length(t)-param.samples_per_ui-5);
    plot(t,PRplt,'r','disp',[param.base 'SBR w Tx/Rx and pkg filters']);
    stem(t(best_cursor_i),best_sbr(best_cursor_i),'g','disp','main cursor sample point');
    title([param.base ' SBR responses'])
    ylabel('volts')
    xlabel('seconds')
    
    stem(sampled_best_sbr_precursors_t, sampled_best_sbr_precursors', 'Color',[.5 .35 .5], 'LineWidth',1,'disp','pre cursors');
    stem(sampled_best_sbr_postcursors_t, sampled_best_sbr_postcursors', 'Color',[.5 .35 .5], 'LineWidth',1,'disp','post cursors');
    stem(sampled_best_sbr_dfecursors_t,dfe_cursors','m', 'LineWidth',2,'disp','dfe cursors');
    grid on
    hold off
end
if OP.DISPLAY_WINDOW
    close(hwaitbar);
else
    fprintf('\n');
end

% % eq_data
result.eq.txle=best_setting;
result.eq.ctle.zero=best_ctle.z;
result.eq.ctle.pole1=best_ctle.p1;
result.eq.ctle.pole2=best_ctle.p2;
result.eq.ctle.maxgain=20*log10(max(abs(ctle_gain)));
result.eq.ctle.maxkacdc=best_ctle.kacdc;
%
result.IR=best_IR;
result.PR=PRin;
result.A_s = best_A_s;
%result.avail_signal=best_cursor;
%result.avail_sig_index=best_cursor_i;
result.best_FOM=best_FOM;

result.DFE_taps = DFE_taps;
result.DFE_taps_i = best_cursor_i+(1:param.ndfe)*param.samples_per_ui;
param.tc=best_cursor_i;
result.sigma_N = best_sigma_N;
result.h_J = best_h_J;



%% TD CTLE function
function [impulse_response p1_ctle, p2_ctle, z_ctle] = TD_CTLE(ir_in,fb,kacdc,oversampling)
fnq=fb/2;
baud_rate=fb;
p1_ctle = -2*pi*fnq/2;
p2_ctle = -2*pi*(2*fnq);
z_ctle = p1_ctle*kacdc;
k_ctle = -p2_ctle;
bilinear_fs = 2*baud_rate*oversampling;
p2d = (1+p2_ctle/bilinear_fs)./(1-p2_ctle/bilinear_fs);
p1d = (1+p1_ctle/bilinear_fs)./(1-p1_ctle/bilinear_fs);
zd = (1+z_ctle/bilinear_fs)./(1-z_ctle/bilinear_fs);
kd = (bilinear_fs-z_ctle)/((bilinear_fs-p1_ctle)*(bilinear_fs-p2_ctle));
B_filt =k_ctle*kd*poly([zd, -1]);
A_filt=poly([p1d, p2d]);
impulse_response=filter(B_filt,A_filt,ir_in);

%% PDF from interference vector
function [ pdf ] = get_pdf_from_sampled_signal( input_vector, L, BinSize )
%   input_vector = list of values of samples
%   return
%   pdf.x
%   pdf.y
%   pdf.vec
%   pdf.bin

if max(input_vector) > BinSize
    input_vector=input_vector(abs(input_vector)>BinSize);
end
% for i = 1:length(input_vector)
%    if abs(input_vector(i)) < BinSize , input_vector(i)=0; end
%end
input_vector(abs(input_vector)<BinSize) = 0;

%% Equation 93A-36 %%
values = 2*(0:L-1)/(L-1)-1;
prob = ones(1,L)/L;

%% Initialize pdf to delta at 0
pdf=d_cpdf(BinSize, 0, 1);

%  if strfind(upper(coding),'PAM4') == 1
%      values=[-abs(input_vector(1)) -abs(input_vector(1))/3 +abs(input_vector(1))/3 +abs(input_vector(1)) ];
%      prob=[.25 .25 .25 .25];
%  else
%      values=[-abs(input_vector(1)) +abs(input_vector(1))];
%      prob=[.5 .5];
%  end

% if length(input_vector) >= 2
for k = 1:length(input_vector)
    %         if strfind(upper(coding),'PAM4') == 1
    %             values=[-abs(input_vector(i)) -abs(input_vector(i))/3 +abs(input_vector(i))/3 +abs(input_vector(i)) ];
    %             prob=[.25 .25 .25 .25];
    %         else
    %             values=[-abs(input_vector(i)) +abs(input_vector(i))];
    %             prob=[.5 .5];
    %         end
    
    pdfn=d_cpdf(BinSize, abs(input_vector(k))*values, prob);
    pdf=conv_fct(pdf, pdfn);
end
% else
%     pdfn.y=1;
%     pdfn.Min=0;
%     pdfn.BinSize=pdf.BinSize;
%     pdf=conv_fct(pdf, pdfn);
% end

function pdf=d_cpdf( binsize, values, probs)
%  p=cpdf(type, ...)
%
% CPDF is actually a PMF for discrete distributions.
% Operations on CPDF objects currently require both objects to have the
% same bin size.
% cpdf(..., 'bin_size', B) sets the bin size to B instead of the default.
%
% cpdf is internally normalized so that the sum of probabilities is 1
% (regardless of bin size).

% Internal fields:
% Min: *bin number* of minimum value.
% BinSize: size of PDF bins. Bin center is the representative value.
% Vec: vector of probabilities per bin.

values=binsize*round(values/binsize);
t=(min(values):binsize:max(values));
pdf.Min=min(values)/binsize;
pdf.y=zeros(size(t));
for k=1:length(values)
    [NA, bin]=min(abs(t-values(k)));
    if isempty(bin)
        warning('no bin');
    end
    pdf.y(bin) = pdf.y(bin)+probs(k);
end

pdf.BinSize=binsize;
pdf.y=pdf.y/sum(pdf.y);
if any(~isreal(pdf.y)) || any(pdf.y<0)
    error('PDF must be real and nonnegative');
end
support=find(pdf.y);
pdf.y=pdf.y(support(1):support(end));
pdf.Min=pdf.Min+(support(1)-1);
pdf.x=(pdf.Min:-pdf.Min)*binsize;

function p=conv_fct(p1, p2)
if p1.BinSize ~= p2.BinSize
    error('bin size must be equal')
end

N1=length(p1.y);
N2=length(p2.y);
p=p1;
p.BinSize=p1.BinSize;
p.Min=p1.Min+p2.Min;
p.y=conv(p1.y, p2.y);
p.x =p.Min*p.BinSize:p.BinSize:-p.Min*p.BinSize;

function pdf = normal_dist(sigma,nsigma,binsize)
pdf.BinSize=binsize;
pdf.Min=-round(nsigma*sigma/binsize);
pdf.x=(pdf.Min:-pdf.Min)*binsize;
pdf.y=exp(-pdf.x.^2/(2*sigma^2+eps));
pdf.y=pdf.y/sum(pdf.y);

function [ param OP ]= read_ParamConfigFile(paramFile, param, OP)
%warning('off','MATLAB:xlsread:Mode'); % suppress warning messages for reading the settings file from XLS
try
    [status,sheets] = xlsfinfo(paramFile);
    if ~isempty(findstr('COM_Settings',cell2mat(sheets)))
        [na1, na2, parameter] = xlsread(paramFile,'COM_Settings','',''); % Import data from the settings file (imports the entire sheet)
    else
        error('COM_Settings tab not in file')
    end
catch ME
    warning('off','MATLAB:xlsread:Mode'); % suppress warning messages for reading the settings file from XLS
    [na1, na2, parameter] = xlsread(paramFile,'','','basic'); % Import data from the settings file (imports the entire sheet)
end
param.type=upper(strtrim(xls_parameter(parameter, 'Coding/Port Type')));
param.fb = xls_parameter(parameter, 'Signal Rate (fb)')*1e9; param.ui=1/param.fb;
param.tx_ffe = -eval(xls_parameter(parameter, '[c(-1) c(1)]'));
param.ndfe = xls_parameter(parameter, 'Nb');
param.max_ctle = -xls_parameter(parameter, 'Gdc, for CTF');
param.a_thru = xls_parameter(parameter, 'Av');
param.a_fext = xls_parameter(parameter, 'Af');
param.a_next = xls_parameter(parameter, 'An');
param.levels = xls_parameter(parameter, 'L');
param.AG=1/(param.levels-1); if param.levels == 2,  param.coding='NRZ'; else param.coding='PAM4'; end
param.specBER = xls_parameter(parameter, 'DER0');
param.cc1 = xls_parameter(parameter, 'CC1');
param.sigma_RJ = xls_parameter(parameter, 'sigma_rj');
param.A_DD = xls_parameter(parameter, 'Add');
param.Na_rms = xls_parameter(parameter, 'sigma_r');
param.eta_0 = xls_parameter(parameter, 'eta_0');
param.samples_per_ui = xls_parameter(parameter, 'Samples Per UI');
param.snpPortsOrder = eval(xls_parameter(parameter, 'Port Order'));
param.ctle_step = xls_parameter(parameter, 'CTF_step');
param.tx_ffe_step = xls_parameter(parameter, 'TXFFE_step');
% This will keep bmax length 0 if Nb=0
param.bmax(1:param.ndfe) = xls_parameter(parameter, 'bmax(1)');
param.bmax(2:param.ndfe) = xls_parameter(parameter, 'bmax(2..Nb)');
param.f_v = xls_parameter(parameter, 'f_v');
param.f_f = xls_parameter(parameter, 'f_f');
param.f_n = xls_parameter(parameter, 'f_n');
param.f_r = xls_parameter(parameter, 'f_r');
param.gamma_il = eval(xls_parameter(parameter, 'package_tl_gamma'));
param.rho_rl = eval(xls_parameter(parameter, 'package_tl_rho'));
param.C_diepad = xls_parameter(parameter, 'C_d')*1e-9; % C_d in nF
param.R_diepad = xls_parameter(parameter, 'R_d');
% allow R_diepad to be a 2-element vector as done in some config files
if ischar(param.R_diepad), param.R_diepad=eval(param.R_diepad); end
switch length(param.R_diepad)
    case 1, param.R_diepad = param.R_diepad*[1 1];
    case 2;
    otherwise
        error('R_diepad should be 1 or 2 element vector');
end

param.C_pkg_board = xls_parameter(parameter, 'C_p')*1e-9; % C_p in nF
param.Pkg_len =  xls_parameter(parameter, 'z_p');
param.WGN_STEP = xls_parameter(parameter, 'WGN_step');
param.BinSize = xls_parameter(parameter, 'PDF_bin_size');
if isempty(strfind(param.type,'CLAUSE 92')) == 0
    param.brd_gamma_il = eval(xls_parameter(parameter, 'board_tl_gamma'));
    param.brd_rho_rl =   eval(xls_parameter(parameter, 'board_tl_rho'));% get operational control variables
    param.brd_len =  xls_parameter(parameter, 'z_bp');
end


% get operational control variables
OP.INCLUDE_CTLE = xls_parameter(parameter, 'INCLUDE_CTLE');
OP.INCLUDE_FILTER = xls_parameter(parameter, 'INCLUDE_TX_RX_FILTER');
OP.DEBUG = xls_parameter(parameter, 'DEBUG');
OP.DISPLAY_WINDOW = xls_parameter(parameter, 'DISPLAY_WINDOW');
OP.CSV_REPORT = xls_parameter(parameter, 'CSV_REPORT');
OP.SAVE_RESP = xls_parameter(parameter, 'SAVE_RESP');
OP.GET_FD = xls_parameter(parameter, 'GET_FD');
OP.INC_PACKAGE = xls_parameter(parameter, 'INC_PACKAGE');
OP.IDEAL_RX_TERM = xls_parameter(parameter, 'IDEAL_RX_TERM');
OP.EXTERNAL = xls_parameter(parameter, 'USE_EXTERNAL_PARAM');
OP.RESULT_DIR = regexprep(xls_parameter(parameter, 'RESULT_DIR'), '\\', filesep);
OP.RX_CALIBRATION = xls_parameter(parameter, 'RX_CALIBRATION');
OP.BREAD_CRUMBS = xls_parameter(parameter, 'BREAD_CRUMBS');
% Parameters computationally defined by values from the settings files
% set up sampling
param.sample_dt = param.ui/param.samples_per_ui;
param.f2 = 1/param.ui;  % used for ICN integrals
param.fb = 1/param.ui;  % used for the filter for ICN
param.sigma_X=sqrt( (param.levels^2-1)/ (3*(param.levels-1)^2) );

function p=xls_parameter(param_sheet, param_name)
% helper function to read parameter values from XLS file. Uses names to find
% values.
[row, col]=find(strcmp(param_sheet, param_name));
if numel(row)*numel(col)==0
    missingParameter(param_name);
elseif numel(row)*numel(col)>1
    % if there are several occurrences, use the first, but warn
    warning('%d occurrences of "%s" found. Using the first', numel(row), param_name);
    row=row(1); col=col(1);
end
p=param_sheet{row, col+1};


function missingParameter (parameterName)
%% FUNCTION :: missingParameter
%
% Description
%   Generates an error message and exits the program due to a parameter in the
%   settings configuration file missing
%
% Input Variables (required)
%   parameterName -- the name of the parameter that has the missing information%
%
% Output/Return Variables
%
% Local Variables (scoped locally to current function only)

error( 'error:badParameterInformation', ...
    '\n\n\t The data for %s is missing or incorrect \n' , parameterName);
%end % missingParameter



function [ H_f  ] = pkg_loss( f1 , param, OP )
f=f1/1e9;

idxf12p9=find(f1>=param.fb/2,1);
param.pkg.a1= 2.9126;
param.pkg.a2= .4835;
param.pkg.a4= .0587;
param.pkg.L_tl= 24.07e-3;
param.pkg.C_d = 2.5e-4;
param.pkg.C_b= 2.2e-4;
param.pkg.vp= .182;
param.pkg.R0= 50;
param.pkg.etl= 11.9;


s21 = exp(-param.pkg.L_tl * ( param.pkg.a1*sqrt(f) + param.pkg.a2*f + param.pkg.a4*f.^2+1i*2*pi*f/param.pkg.vp ));
s12=s21;

G_die =  -1i*f./( 1./(pi*param.pkg.C_d*param.pkg.R0)+1i.*f );
G_ball = -1i*f./( 1./(pi*param.pkg.C_b*param.pkg.R0)+1i.*f );


s11 = param.pkg.etl/(200+param.pkg.etl);
s22 = s11;
%H_f = Pure_Loss ./ (1 - (G_die.* param.pkg.Gtl + G_ball.*param.pkg.Gtl + G_die.*G_ball.*Pure_Loss.^2 - G_die.*G_ball.*param.pkg.Gtl.*param.pkg.Gtl));
H_f = s21./(1-s11.*G_die-s22*G_ball+G_die.*G_ball.*(s11.*s22-s21.*s12));
if OP.DEBUG
    figure(305);
    ilatnq=num2str(20*log10(abs(H_f(idxf12p9))));
    subplot(3,1,1); plot(f, 20*log10(abs(s21)), f, 20*log10(abs(G_die)), f, 20*log10(abs(G_ball)));
    subplot(3,1,2); plot(f, 20*log10(abs(H_f)),'disp',['IL at Fnq= ' ilatnq]);
    legend show
    subplot(3,1,3); plot(f, unwrap(angle(s21))/pi*180, f, unwrap(angle(H_f))/pi*180);
end


function [c] = Par(a,b)
c=a.*b./(a+b);



function [c] = par(a,b)
c=a.*b./(a+b);

function csvwrite_with_headers(filename,m,headers,r,c)

% This function functions like the build in MATLAB function csvwrite but
% allows a row of headers to be easily inserted
%
% known limitations
% 	The same limitation that apply to the data structure that exist with
%   csvwrite apply in this function, notably:
%       m must not be a cell array
%
% Inputs
%
%   filename    - Output filename
%   m           - array of data
%   headers     - a cell array of strings containing the column headers.
%                 The length must be the same as the number of columns in m.
%   r           - row offset of the data (optional parameter)
%   c           - column offset of the data (optional parameter)
%
%
% Outputs
%   None

%% initial checks on the inputs
if ~ischar(filename)
    error('FILENAME must be a string');
end

% the r and c inputs are optional and need to be filled in if they are
% missing
if nargin < 4
    r = 0;
end
if nargin < 5
    c = 0;
end

if ~iscellstr(headers)
    error('Header must be cell array of strings')
end


if length(headers) ~= size(m,2)
    error('number of header entries must match the number of columns in the data')
end

%% write the header string to the file

%turn the headers into a single comma separated string if it is a cell
%array,
header_string = headers{1};
for i = 2:length(headers)
    header_string = [header_string,',',headers{i}];
end
%if the data has an offset shifting it right then blank commas must
%be inserted to match
if r>0
    for i=1:r
        header_string = [',',header_string];
    end
end

%write the string to a file
fid = fopen(filename,'w');
fprintf(fid,'%s\r\n',header_string);
fclose(fid);

%% write the append the data to the file

%
% Call dlmwrite with a comma as the delimiter
%
dlmwrite(filename, m,'-append','delimiter',',','roffset', r,'coffset',c);

function [sdd21p] = s21_pkg(sdd21, sdd11, sdd22, faxis, param, OP, channel_type, channel_number)
% concatenates package reflections with s21,s11,and s22 with spec return loss (gammas)
% faxis is the frequency array
% sdd21, sdd11, sdd22 are the corresponding array of differential parameters
Z0=50;

C_diepad = param.C_diepad;
C_pkg_board = param.C_pkg_board;
if isequal(channel_type ,'NEXT')
    C_diepad = C_diepad/2;
    C_pkg_board = C_pkg_board/2;
end
[ s11out, s12out, s21out, s22out ]= make_pkg( faxis, param.rho_rl, param.gamma_il, param.Pkg_len, C_diepad, C_pkg_board );
gamma_tx=(param.R_diepad(1)-Z0)/(param.R_diepad(1)+Z0);% equation 93A-15
if OP.IDEAL_RX_TERM == 1
    gamma_rx=0; 
else
    gamma_rx=(param.R_diepad-Z0)/(param.R_diepad+Z0);% equation 93A-15
end

if OP.INC_PACKAGE==0
    sdd21p= sdd21;
else
    if OP.RX_CALIBRATION == 1 && channel_number == 2
        %   for calibration do not include the transmitter package
        [s11out_rx, s12out_rx, s21out_rx, s22out_rx ] = combines4p( sdd11, sdd21, sdd21, sdd22, s22out, s12out, s21out, s11out ); % s22 is ball side of package %% equation 93A-14
        %% Equation 93A-16   
        sdd21p= s21out_rx.*(1-gamma_tx).*(1+gamma_rx)./(1.- s11out_rx.*gamma_tx - s22out_rx.*gamma_rx  -s21out_rx.^2.*gamma_tx.*gamma_rx +s11out_rx.*s22out_rx.*gamma_tx.*gamma_rx);
    else
        % set s21 for package to 1 to compensate for spec guaranteed minimum voltage swing
        %% Equations 93A-4 to 93A-7   
        if OP.IDEAL_RX_TERM == 1
            [s11out_rx, s12out_rx, s21out_rx, s22out_rx ] = combines4p(   s11out, s12out, s21out, s22out, sdd11, sdd21, sdd21, sdd22 );
        else
            [s11out_txdie, s12out_ball_die, s21out_die_ball, s22out_rxball ] = combines4p(   s11out, s12out, s21out, s22out, sdd11, sdd21, sdd21, sdd22 );
            [s11out_rx, s12out_rx, s21out_rx, s22out_rx ] = combines4p( s11out_txdie, s12out_ball_die, s21out_die_ball, s22out_rxball  , s22out, s12out, s21out, s11out ); % s22 is ball side of package
        end
        %% Equation 93A-16
        sdd21p= s21out_rx.*(1-gamma_tx).*(1+gamma_rx)./(1.- s11out_rx.*gamma_tx - s22out_rx.*gamma_rx  -s21out_rx.^2.*gamma_tx.*gamma_rx +s11out_rx.*s22out_rx.*gamma_tx.*gamma_rx);
    end
end


function [ s11out, s12out, s21out, s22out ] = make_pkg( f, rho_rl,gamma_il, Lenscale, cpad, cball )
zref=50;
% termination impedance
f(f<eps)=eps;
% zpad =1./(2*1i*pi.*f*cpad);
% s11pad = -zref./(zref+2*zpad);
% s21pad = 2*zpad./(zref+2*zpad);
%% Equation 93A-8
s11pad= -1i*2*pi.*f*cpad*zref./(2+1i*2*pi.*f*cpad*zref);
s21pad= 2./(2+1i*2*pi.*f*cpad*zref);
S21 = synthesize_tline(Lenscale,f, gamma_il);% equation 93A-9
S11 = scales11(Lenscale,f, gamma_il, rho_rl);% equation 93A-10
[ s11out1, s12out1, s21out1, s22out1 ]=combines4p(  s11pad, s21pad, s21pad, s11pad, S11, S21, S21, S11 ); % first part of equation 93A-13

% zball = 1./(2*1i*pi.*f*cball); % Add the ball discontinuity
% s11ball = -zref./(zref+2*zball);
% s21ball = 2*zball./(zref+2*zball);
%% Equation 93A-8
s11ball= -1i*2*pi.*f*cball*zref./(2+1i*2*pi.*f*cball*zref);
s21ball= 2./(2+1i*2*pi.*f*cball*zref);
[ s11out, s12out, s21out, s22out ]=combines4p( s11out1, s12out1, s21out1, s22out1, s11ball, s21ball, s21ball, s11ball );% second part of equation 93A-13



%% for Clause 92
function [ s11out, s12out, s21out, s22out ] = add_brd(chdata, i, param)
[ s11tx, s12tx, s21tx, s22tx ] = make_trace( chdata(i).faxis, param.brd_rho_rl, param.brd_gamma_il, param.brd_len);
[ s11rx, s12rx, s21rx, s22rx ] = make_trace( chdata(i).faxis, param.brd_rho_rl, param.brd_gamma_il, param.brd_len);
[ s11out1, s12out1, s21out1, s22out1 ]=combines4p(  s11tx, s12tx, s21tx, s22tx, chdata(i).sdd11, chdata(i).sdd21, chdata(i).sdd12, chdata(i).sdd22 ); 
[ s11out, s12out, s21out, s22out ]=combines4p(  s11out1, s12out1, s21out1, s22out1,  s11rx, s12rx, s21rx, s22rx); 

function [ s11out, s12out, s21out, s22out ] = make_trace( f, rho_rl,gamma_il, Lenscale)
s12out = synthesize_tline(Lenscale,f, gamma_il);%% equation 93A-9
s11out = scales11(Lenscale,f, gamma_il, rho_rl);%% equation 93A-10
s22out=s11out;
s21out=s12out;
%%


function FIT = synthesize_tline(lenscale, f, gamma_il)
efit=polyval(lenscale*fliplr(gamma_il), sqrt(f/1e9));
FIT=transpose(exp(transpose(efit)));

function [s11new] = scales11(lenscale, f, gamma_il, rho_rl )
%UNTITLED Summary of this function goes here
%   Detailed explanation goes here
s11new=1;
%s11fit=alpha_rl(1)+alpha_rl(2).*sqrt(f)+alpha_rl(3).*f +f.^2.*alpha_rl(4);
%S11base=transpose(exp(transpose(s11fit)));
S11base=polyval(fliplr(rho_rl), f/1e9); % Comment #101 on D2.0; benartsi_3bj_01a_0513
%s21fit=gamma_il(1)+gamma_il(2).*sqrt(f)+gamma_il(3).*f +f.^2.*gamma_il(4);
%S21base=transpose(exp(transpose(s21fit)));
S21base=synthesize_tline(1, f, gamma_il);
for k=2:2:2*lenscale-2
    s11new=s11new+S21base.^k;
end
s11new=s11new.*S11base;

function [ s11out, s12out, s21out, s22out ] = combines4p( s11in1, s12in1, s21in1, s22in1, s11in2, s12in2, s21in2, s22in2)
%UNTITLED Summary of this function goes here
%   Detailed explanation goes here
for i=1:length(s11in1) s1(:,:,i)=[s11in1(i) s12in1(i); s21in1(i) s22in1(i) ]; end
for i=1:length(s11in1) s2(:,:,i)=[s11in2(i) s12in2(i); s21in2(i) s22in2(i) ]; end
t1=stot(s1);
t2=stot(s2);
for i=1:length(s11in1) t3(:,:,i)=t1(:,:,i)*t2(:,:,i); end
s3=ttos(t3); %
s11out=s3(1,1,:);
s11out=transpose(s11out(:));
s12out=s3(1,2,:);
s12out=transpose(s12out(:));
s21out=s3(2,1,:);
s21out=transpose(s21out(:));
s22out=s3(2,2,:);
s22out=transpose(s22out(:));



function t_params = stot(s_params)
% p 67 R. Mavaddat. (1996). Network scattering parameter. Singapore: World Scientific.
% ISBN 978-981-02-2305-2. http://books.google.com/?id=287g2NkRYxUC&lpg=PA65&dq=T-parameters+&pg=PA67.
[s11, s12, s21, s22] = deal(s_params(1,1,:), s_params(1,2,:), s_params(2,1,:), s_params(2,2,:));
delta = (s11.*s22-s12.*s21);
t_params = [1./s21, -s22./s21; s11./s21, -delta./s21];


function s_params = ttos(t_params)
% p 67 R. Mavaddat. (1996). Network scattering parameter. Singapore: World Scientific.
% ISBN 978-981-02-2305-2. http://books.google.com/?id=287g2NkRYxUC&lpg=PA65&dq=T-parameters+&pg=PA67.
[t11, t12, t21, t22] = deal(t_params(1,1,:), t_params(1,2,:), t_params(2,1,:), t_params(2,2,:));
delta = t11.*t22-t21.*t12;
s_params = [t21./t11, delta./t11; 1./t11, -t12./t11];

function [ sigma_noise sigma_r_mod ] = get_sigma_noise( kacdc, param, chdata )
% the FEXT channel for calibration basically a DC connection unlike normal
% FEXT channels which are nearly open at DC channels
if size(chdata,2) >= 2
    idxfbby2=find( chdata(2).faxis(:) >= param.fb/2, 1);
    
    f=chdata(2).faxis(1:idxfbby2);
    Hnoise_channel=abs(chdata(2).sdd21(1:idxfbby2));
else
    f=0:10e6:param.fb/2;
    Hnoise_channel=1;
end
hctl=param.fb*(1i.*f+0.25*param.fb*kacdc)./((1i*f+0.25*param.fb).*(1i*f+param.fb));
bw1=2.613126; bw2=3.4142136; bw3=2.613126; fr=param.f_r*param.fb;
Hrx=(1+bw1*1i*f./fr+bw2*(1i*f./fr).^2+bw3*(1i*f./fr).^3+(1i*f./fr).^4).^-1;
H_np=Hnoise_channel.*hctl.*Hrx;
sigma_noise = sqrt((sum(abs(H_np)).^2))/length(H_np) * param.WGN;
sigma_r_mod= sqrt(param.Na_rms^2+sigma_noise^2);

function [ sigma_XT ] = get_xtlk_noise( kacdc,upsampled_txffe, param, chdata )
% resample xtlk for each sample N is num crosstalk and M is phase (samp/UI)
% k is index for crosstalk
% n is UI index
% m is sample index
K=param.K;
M=param.samples_per_ui;
for k=2:K
    sigma_i_sqd(k)=0;
    effective_channel = filter(upsampled_txffe, 1, chdata(k).impulse_response);
    N=round(length(chdata(k).impulse_response)/param.samples_per_ui)-3;
    sbr=filter(ones(param.samples_per_ui, 1), 1, effective_channel);
    for m=1:M
        h=sbr(param.samples_per_ui*(1:N)+m);
        sigma_i_sqd(k)=max(param.sigma_X^2*sum(h.^2),sigma_i_sqd(k));
    end
end
sigma_XT_sqd=0;
for k=2:K
    sigma_XT_sqd = sigma_XT_sqd + sum(sigma_i_sqd(k));
end
sigma_XT=sqrt(sigma_XT_sqd);
%----------------------- Manual version change log ----------------------
% Version 1.0 08/14/2012
% Version 1.01 08/17/2012 updating to draft and adding more parameter to xls sheet
% Version 1.02 08/31/2012 fix parsing for PAM4, added maxdfex code.
% Verson  1.05 has new vtf and apply to all s parameters IEEE posted
% version d1p2 added package IL and package RL per draft 1.2
%              pkg parameter removed and set by xls
%              noise included in h_ISI region for pam4 base on param.bmax
%              precuror ISI include in PDF as well as noise in DFE region
%              Noise updated to Draft 1.2 spec in FOM calculation
%              now XLS input is always required.
%              package ref frequency is now absolute
% version d1p2a fixed sweeping equalization cursor finding bug that occur if
%              eq pare meters are altered in xls sheet
% version d1p21b name change to reflect doc file
% version d1p21d including precusor isi
%                isi for interference signal is only at cursor+N*UI sample
%                points
% version post_d1p2 report dfe coef in debug
% version pre_d1p3a vtv
% version post_d1p3 added new package model and include a few fixes in
%                    sweep_eq
% version post_d1p3a new pkg modeling and calibration operation
% version post_d1p3b OP.BMXLMT=1; % set to 1 to limit dfe tap 1 to 1
% version post_d1p3b_adee Added CTLE effect to AWGN; corrected inverted logic for
%                         OP.BMXLMT in one place
% version post_d1p3c Merging with changes in post_d1p3a1 which are:
%                     i)corrected error from pre-allocation of frequency index in function read_p4_s4params
%                       %Search: freqCounter
%                     ii)program will now error if number of samples does not
%                     meet expected range of Nb
%                       %Search: length(sbr) < cursor_i+param.samples_per_ui*(param.ndfe+1)
%                     iii)added informative warning to user if max frequency and frequency step
%                     are less than signaling rate and delta f

% $Log: com_ieee8023_93a.m.rca $
% 
%  Revision: 1.17 Thu Jun 20 10:19:39 2013 aran
%  Implemented changes in feedback from Stefaan Sercu. Removed instances of ~ in call results for backward compatibility.
% 
%  Revision: 1.16 Tue Jun 18 23:30:05 2013 aran
%  Added support for Clause 92 (comment 188); additional small fixes
% 
%  Revision: 1.14 Thu Jun 13 01:05:57 2013 aran
%  Implementation of updated jitter calculations done (pre validation)
%  Added references to equation numbers, changed some variable names to align with D2.1
%
%  Revision: 1.13 Mon Jun 10 15:16:14 2013 aran
%  merged more changes by Rich. corrected sigma_N (sigma_r_mod) calculations.
%
%  Revision: 1.12 Thu Jun  6 16:18:39 2013 aran
%  Merged changes by Rich Mellitz
%
%  Revision: 1.11 Wed Jun  5 18:29:08 2013 aran
%  Implemented comment #80 on D2.0 - b_max is a vector, first element can be specified separately; Includes support for ndfe=0
%
%  Revision: 1.10 Mon Jun  3 00:36:33 2013 aran
%  Implemented comment #73 on D2.0: sigma_r calculated from noise spectral density.
%  CTLE is applied to noise.
%  Added PDF_bin_size parameter.
%
%  Revision: 1.8 Tue May 28 20:51:47 2013 aran
%  implemented comment #155 on D2.0: reflection coefficients negated
%
%  Revision: 1.7 Tue May 28 20:40:41 2013 aran
%  completed implementing comment #101 on D2.0; changed XLS reading code - now based on text labels in the file
%
%  Revision: 1.6 Tue May 28 11:34:42 2013 aran
%  implemented comment #101 on D2.0: RL fitting
%
%  Revision: 1.5 Sun May 26 23:49:00 2013 aran
%  Updated graphics, proof-read text, log moved to end
